import os
import shutil
import tempfile
import unittest

import cwepr.dataset
import cwepr.io

ROOTPATH = os.path.split(os.path.abspath(__file__))[0]


class TestBES3TImporter(unittest.TestCase):
    def setUp(self):
        self.dataset = cwepr.dataset.ExperimentalDataset()

    def test_import_with_1D_dataset(self):
        source = os.path.join(ROOTPATH, 'testdata/test-bes3t-1D-fieldsweep.DSC')
        importer = cwepr.io.bes3t.BES3TImporter(source=source)
        self.dataset.import_from(importer)
        self.assertTrue(self.dataset.data.axes[0].unit)
        self.assertFalse(self.dataset.data.axes[1].unit)

    def test_import_with_1D_dataset_dimensions(self):
        source = os.path.join(ROOTPATH, 'testdata/test-bes3t-1D-fieldsweep.DSC')
        importer = cwepr.io.bes3t.BES3TImporter(source=source)
        self.dataset.import_from(importer)
        self.assertEqual(2, len(self.dataset.data.axes))
        self.assertEqual(1, self.dataset.data.data.ndim)

    def test_import_gives_correct_units(self):
        source = os.path.join(ROOTPATH, 'testdata/BDPA-1DFieldSweep')
        importer = cwepr.io.bes3t.BES3TImporter(source=source)
        self.dataset.import_from(importer)
        self.assertAlmostEqual(0.6325,
                         self.dataset.metadata.bridge.power.value)
        self.assertEqual('mW',
                         self.dataset.metadata.bridge.power.unit)

        self.assertEqual('mT', self.dataset.metadata.magnetic_field.start.unit)
        self.assertEqual('mT',
                         self.dataset.metadata.magnetic_field.sweep_width.unit)
        self.assertTrue(self.dataset.data.axes[0].values[0] < 1000)

    def test_import_with_2D_dataset_powersweep(self):
        source = os.path.join(ROOTPATH, 'testdata/BDPA-2DFieldPower.DSC')
        importer = cwepr.io.bes3t.BES3TImporter(source=source)
        self.dataset.import_from(importer)
        self.assertTrue(self.dataset.data.axes[0].unit)
        self.assertTrue(self.dataset.data.axes[1].unit)
        self.assertTrue(self.dataset.data.axes[1].quantity)

    def test_import_with_2D_dataset_fielddelay(self):
        source = os.path.join(ROOTPATH, 'testdata/BDPA-2DFieldDelay.DSC')
        importer = cwepr.io.bes3t.BES3TImporter(source=source)
        self.dataset.import_from(importer)
        self.assertTrue(self.dataset.data.axes[0].unit)
        self.assertTrue(self.dataset.data.axes[1].unit)

    def test_imports_infofile(self):
        source = os.path.join(ROOTPATH, 'testdata/BDPA-2DFieldDelay.DSC')
        importer = cwepr.io.bes3t.BES3TImporter(source=source)
        self.dataset.import_from(importer)
        self.assertTrue(self.dataset.metadata.measurement.operator)
        self.assertTrue(self.dataset.metadata.measurement.purpose)
        self.assertTrue(self.dataset.metadata.experiment.type)
        self.assertTrue(self.dataset.metadata.probehead.type)
        self.assertTrue(self.dataset.metadata.temperature_control.temperature
                        .value)

    def test_import_with_no_infofile_continues(self):
        source = os.path.join(ROOTPATH, 'testdata/BDPA-2DFieldDelay')
        with tempfile.TemporaryDirectory() as testdir:
            for extension in ('.DSC', '.DTA', '.YGF'):
                new_source = os.path.join(testdir, 'test-wo-infofile')
                shutil.copyfile(source + extension, new_source + extension)
            dataset = cwepr.dataset.ExperimentalDataset()
            importer = cwepr.io.bes3t.BES3TImporter(
                source=new_source)
            dataset.import_from(importer)



"""
Plotting: Graphical representations of data extracted from datasets.

Graphical representations of cw-EPR data are an indispensable aspect of data
analysis. To facilitate this, a series of different plotters are available.

Plotting relies on `matplotlib <https://matplotlib.org/>`_, and mainly its
object-oriented interface should be used for the actual plotting.

Generally, two types of plotters can be distinguished:

* Plotters for handling single datasets

  Shall be derived from :class:`aspecd.plotting.SinglePlotter`.

* Plotters for handling multiple datasets

  Shall be derived from :class:`aspecd.plotting.MultiPlotter`.

In the first case, the plot is usually handled using the :meth:`plot` method
of the respective :obj:`cwepr.dataset.Dataset` object. Additionally,
those plotters always only operate on the data of a single dataset, and the
plot can easily be attached as a representation to the respective dataset.
Plotters handling single datasets should always inherit from the
:class:`aspecd.plotting.SinglePlotter` class.

In the second case, the plot is handled using the :meth:`plot` method of the
:obj:`aspecd.plotting.Plotter` object, and the datasets are stored as a list
within the plotter. As these plots span several datasets, there is no easy
connection between a single dataset and such a plot in sense of
representations stored in datasets. Plotters handling multiple datasets should
always inherit from the :class:`aspecd.plotting.MultiPlotter` class.


A note on array dimensions and axes
===================================

Something often quite confusing is the apparent inconsistency between the
order of array dimensions and the order of axes. While we are used to assign
axes in the order *x*, *y*, *z*, and assuming *x* to be horizontal,
*y* vertical (and *z* sticking out of the paper plane), arrays are usually
indexed row-first, column-second. That means, however, that if you simply
plot a 2D array in axes, your *first* dimension is along the *y* axis,
the *second* dimension along the *x* axis.

Therefore, as the axes of your datasets will always correspond to the array
dimensions of your data, in case of 2D plots you will need to *either* use
the information contained in the second axis object for your *x* axis label,
and the information from the first axis object for your *y* axis label,
*or* to transpose the data array.

Another aspect to have in mind is the position of the origin. Usually,
in a Cartesian coordinate system, convention is to have the origin (0,
0) in the *lower left* of the axes (for the positive quadrant). However,
for images, convention is to have the corresponding (0, 0) pixel located in
the *upper left* edge of your image. Therefore, those plotting methods
dealing with images will usually *revert* the direction of your *y* axis.
Most probably, eventually you will have to check with real data and ensure
the plotters to plot data and axes in a consistent fashion.


Types of concrete plotters
==========================

The cwepr package comes with a series of concrete plotters included ready
to be used, thanks to inheriting from the underlying ASpecD framework. As
stated above, plotters can generally be divided into two types: plotters
operating on single datasets and plotters combining the data of multiple
datasets into a single figure.

Additionally, plotters can be categorised with regard to creating figures
consisting of a single or multiple axes. The latter are plotters inheriting
from the :class:`aspecd.plotting.CompositePlotter` class. The latter can be
thought of as templates for the other plotters to operate on, *i.e.* they
provide the axes for other plotters to display their results.


Concrete plotters for single datasets
-------------------------------------

* :class:`cwepr.plotting.SinglePlotter1D`

  Basic line plots for single datasets, allowing to plot a series of
  line-type plots, including (semi)log plots

* :class:`cwepr.plotting.SinglePlotter2D`

  Basic 2D plots for single datasets, allowing to plot a series of 2D plots,
  including contour plots and image-type display

* :class:`cwepr.plotting.SingleCompositePlotter`

  Composite plotter for single datasets, allowing to plot different views of
  one and the same datasets by using existing plotters for single datasets.

* :class:`cwepr.plotting.GoniometerSweepPlotter`

  Composite plotter for single datasets representing goniometer sweeps,
  *i.e.* angular-dependent cw-EPR measurements.


Concrete plotters for multiple datasets
---------------------------------------

* :class:`cwepr.plotting.MultiPlotter1D`

  Basic line plots for multiple datasets, allowing to plot a series of
  line-type plots, including (semi)log plots

* :class:`cwepr.plotting.MultiPlotter1DStacked`

  Stacked line plots for multiple datasets, allowing to plot a series of
  line-type plots, including (semi)log plots


Module documentation
====================

"""
import copy

import aspecd.plotting
import aspecd.processing


class GoniometerSweepPlotter(aspecd.plotting.SingleCompositePlotter):
    """Overview of the results of a goniometer sweep.

    A goniometer sweep, *i.e.* a series of cw-EPR spectra as a function of
    the angle of the sample with respect to the external magnetic field,
    is usually performed over at least 180°, regardless of the step size.
    The reason is simply that the spectra for 0° and 180° should be
    identical due to the underlying physics of magnetic resonance.

    The plotter will create three subpanels:

    * A 2D plot (scaled image plot) as a general overview.

    * A 1D multiplot comparing the signals for 0° and 180° to check for
      consistency during the measurement.

    * A stacked plot showing all angular positions, providing an alternative
      view of the angular-dependent signal changes compared to the 2D plot.


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class.

    To get an overview of your goniometer sweep, just invoke the plotter with
    default values:

    .. code-block:: yaml

       - kind: singleplot
         type: GoniometerSweepPlotter
         properties:
           filename: output.pdf

    """

    def __init__(self):
        super().__init__()
        self.description = 'Plot for one goniometric dataset in different ' \
                           'representations.'
        self.grid_dimensions = [2, 2]
        self.subplot_locations = [[0, 0, 1, 1], [1, 0, 1, 1], [0, 1, 2, 1]]
        self.plotter = [aspecd.plotting.SinglePlotter2D(),
                        aspecd.plotting.MultiPlotter1D(),
                        aspecd.plotting.SinglePlotter2DStacked()]
        self.axes_positions = [[0, 0.15, 1, 1], [0, 0, 1, 1],
                               [0.25, 0, 0.9, 1.07]]
        self.zero_deg_slice = None
        self.hundredeighty_deg_slice = None

    def _set_properties(self):
        upper_contour = self.plotter[0]
        upper_contour.type = 'contourf'
        upper_contour.parameters['show_contour_lines'] = True

    def _create_plot(self):
        self._configure_contour_plotter()
        self._extract_traces()
        self._configure_comparison_plotter()
        self._set_properties()
        super()._create_plot()

    def _extract_traces(self):
        slicing = aspecd.processing.SliceExtraction()
        slicing.parameters['position'] = 0
        slicing.parameters['unit'] = 'axis'
        slicing.parameters['axis'] = 1
        self.zero_deg_slice = copy.deepcopy(self.dataset)
        self.zero_deg_slice.process(slicing)
        self.zero_deg_slice.label = '0°'
        slicing.parameters['position'] = 180
        self.hundredeighty_deg_slice = copy.deepcopy(self.dataset)
        self.hundredeighty_deg_slice.process(slicing)
        self.hundredeighty_deg_slice.label = '180°'

    def _configure_comparison_plotter(self):
        comparison_plotter = aspecd.plotting.MultiPlotter1D()
        comparison_plotter.datasets = [self.zero_deg_slice,
                                       self.hundredeighty_deg_slice]
        comparison_plotter.properties.from_dict({
            'drawings': [
                {'color': 'tab:blue'},
                {'color': 'tab:red'}
            ],
            'axes': {
                'yticks': [],
                'ylabel': r'$EPR\ intensity$'
            }
        })
        comparison_plotter.parameters['show_legend'] = True
        self.plotter[1] = comparison_plotter

    def _configure_contour_plotter(self):
        self.plotter[0].properties.from_dict({
            'axes': {
                'yticks': [0, 30, 60, 90, 120, 150, 180]
            }
        })


class SinglePlotter1D(aspecd.plotting.SinglePlotter1D):
    """1D plots of single datasets.

    Convenience class taking care of 1D plots of single datasets.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.plotting.SinglePlotter1D`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. Of course, all parameters settable
    for the superclasses can be set as well. The examples focus each on a
    single aspect.

    In the simplest case, just invoke the plotter with default values:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter1D
         properties:
           filename: output.pdf

    """


class SinglePlotter2D(aspecd.plotting.SinglePlotter2D):
    """2D plots of single datasets.

    Convenience class taking care of 2D plots of single datasets.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.plotting.SinglePlotter2D`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. Of course, all parameters settable
    for the superclasses can be set as well. The examples focus each on a
    single aspect.

    In the simplest case, just invoke the plotter with default values:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2D
         properties:
           filename: output.pdf

    To change the axes (flip *x* and *y* axis):

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2D
         properties:
           filename: output.pdf
           parameters:
             switch_axes: True

    To use another type (here: contour):

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2D
         properties:
           filename: output.pdf
           type: contour

    To set the number of levels of a contour plot to 10:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2D
         properties:
           filename: output.pdf
           type: contour
           parameters:
             levels: 10

    To change the colormap (cmap) used:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2D
         properties:
           filename: output.pdf
           properties:
             drawing:
               cmap: RdGy

    Make sure to check the documentation of the ASpecD
    :mod:`aspecd.plotting` module for further parameters that can be set.

    """


class MultiPlotter1D(aspecd.plotting.MultiPlotter1D):
    """1D plots of multiple datasets.

    Convenience class taking care of 1D plots of multiple datasets.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.plotting.MultiPlotter1D`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. Of course, all parameters settable
    for the superclasses can be set as well. The examples focus each on a
    single aspect.

    In the simplest case, just invoke the plotter with default values:

    .. code-block:: yaml

       - kind: multiplot
         type: MultiPlotter1D
         properties:
           filename: output.pdf

    To change the settings of each individual line (here the colour and label),
    supposing you have three lines, you need to specify the properties in a
    list for each of the drawings:

    .. code-block:: yaml

       - kind: multiplot
         type: MultiPlotter1D
         properties:
           filename: output.pdf
           properties:
             drawings:
               - color: '#FF0000'
                 label: foo
               - color: '#00FF00'
                 label: bar
               - color: '#0000FF'
                 label: foobar

    .. important::
        If you set colours using the hexadecimal RGB triple prefixed by
        ``#``, you need to explicitly tell YAML that these are strings,
        surrounding the values by quotation marks.

    """


class MultiPlotter1DStacked(aspecd.plotting.MultiPlotter1DStacked):
    """Stacked 1D plots of multiple datasets.

    Convenience class taking care of 1D plots of multiple datasets.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.plotting.MultiPlotter1DStacked`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. Of course, all parameters settable
    for the superclasses can be set as well. The examples focus each on a
    single aspect.

    In the simplest case, just invoke the plotter with default values:

    .. code-block:: yaml

       - kind: multiplot
         type: MultiPlotter1DStacked
         properties:
           filename: output.pdf

    To change the settings of each individual line (here the colour and label),
    supposing you have three lines, you need to specify the properties in a
    list for each of the drawings:

    .. code-block:: yaml

       - kind: multiplot
         type: MultiPlotter1DStacked
         properties:
           filename: output.pdf
           properties:
             drawings:
               - color: '#FF0000'
                 label: foo
               - color: '#00FF00'
                 label: bar
               - color: '#0000FF'
                 label: foobar

    .. important::
        If you set colours using the hexadecimal RGB triple prefixed by
        ``#``, you need to explicitly tell YAML that these are strings,
        surrounding the values by quotation marks.

    Sometimes you want to have horizontal "zero lines" appear for each
    individual trace of the stacked plot. This can be achieved explicitly
    setting the "show_zero_lines" parameter to "True" that is set to "False"
    by default:

    .. code-block:: yaml

       - kind: multiplot
         type: MultiPlotter1DStacked
         properties:
           filename: output.pdf
           parameters:
             show_zero_lines: True

    """


class CompositePlotter(aspecd.plotting.CompositePlotter):
    """Base class for plots consisting of multiple axes.

    The underlying idea of composite plotters is to use a dedicated
    existing plotter for each axis and assign this plotter to the list of
    plotters of the CompositePlotter object. Thus the actual plotting task
    is left to the individual plotter and the CompositePlotter only takes
    care of the specifics of plots consisting of more than one axis.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.plotting.CompositePlotter`
    class for details.

    """


class SingleCompositePlotter(aspecd.plotting.SingleCompositePlotter):
    """Composite plotter for single datasets.

    This composite plotter is used for different representations of one and the
    same dataset in multiple axes contained in one figure. In this respect,
    it works like all the other ordinary single plotters derived from
    :class:`SinglePlotter`, *i.e.* it usually gets called by using the dataset's
    :meth:`aspecd.dataset.Dataset.plot` method.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.plotting.SingleCompositePlotter`
    class for details.

    """


class Saver(aspecd.plotting.Saver):
    """Base class for saving plots.

    For basic saving of plots, no subclassing is necessary, as the
    :meth:`save` method uses :meth:`matplotlib.figure.Figure.savefig` and
    can cope with all possible parameters via the :attr:`parameters` property.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.plotting.Saver` class for
    details.

    """



"""Report implementation for cwepr module.

.. note::
    The dataset can be given either as dataset in the properties of a recipe, or
    via the apply_to parameter. In the first case, the dataset can be accessed
    in the here implemented reporter class via ``self.dataset`` (and as an
    object), in the latter
    case indirectly via operating on the context-object ``self.context[
    'dataset']`` (and as a dict). The dataset has to be given explicitly
    while the dataset-context is applied implicitly. Therefore, here is
    applied the usage of the context that is a bit more complicated in
    operating but more intuitive to write in recipes.


.. note::
    Still in active developing and not fail safe and easy to use.
"""

import collections
import copy
import os

import aspecd.report
import aspecd.dataset

import cwepr.dataset


class ExperimentalDatasetLaTeXReporter(aspecd.report.LaTeXReporter):
    """Report implementation for cwepr module."""

    def __init__(self, template='', filename=''):
        super().__init__(template=template, filename=filename)
        self.dataset = cwepr.dataset.ExperimentalDataset()
        # private properties
        self._metadata = dict()
        self._tasks = collections.OrderedDict()
        self._figure_name = dict()

    def create(self):
        """Perform all methods to generate a report."""
        self._prepare_metadata()
        self._get_tasks()
        self._get_figure_names()
        self._create_context()
        self.context = self._sanitise_context(self.context)
        super().create()

    def _prepare_metadata(self):
        self._metadata = self.context['dataset']['metadata']
        self._metadata['parameter'] = collections.OrderedDict()
        self._collect_experimental_parameters()

    def _collect_experimental_parameters(self):
        """Collect all the metadata keys."""
        for key in self._metadata.keys():
            if key not in ['sample', 'measurement', 'parameter']:
                self._metadata['parameter'][key] = \
                    self._metadata[key]

    def _get_tasks(self):
        for task in self.dataset.tasks:
            if task['kind'] in ('analysis', 'processing'):
                self._tasks[(getattr(task['task'],
                                     task['kind']).description)] = {
                    'Parameters': getattr(task['task'],
                                          task['kind']).parameters,
                    'Comment': getattr(task['task'], task['kind']).comment
                }

    def _get_tasks_recursively(self, dict_=None):
        """Gets tasks recursively.

        Is able to get all the tasks performed on an single dataset what is
        only possible thanks to *call by reference* of the history records (
        ?). The dataset initially containing all the data has to be given in
        the recipe explicitly.

        .. note::
            Currently not used

        .. note: :
            Currently, the order of the tasks is *not* preserved because of
            the recursive search and the third step appears before the
            second. Therefore rework this with a different method, probably
            an other *context* keyword in a recipe.

        """
        for task in dict_.tasks:
            if task['kind'] == 'annotation':
                continue
            if task['kind'] == 'analysis':
                if isinstance(task['task'].analysis.result,
                              aspecd.dataset.CalculatedDataset):
                    self._get_tasks_recursively(task['task'].analysis.result)

            self._tasks[(getattr(task['task'],
                                 task['kind']).description)] = {
                'Parameters': getattr(task['task'],
                                      task['kind']).parameters,
                'Comment': getattr(task['task'], task['kind']).comment,
            }

    def _create_context(self):
        """Create a dictionary containing all data to write the report."""
        self.context['TASKS'] = self._tasks
        self.context['METADATA'] = self._metadata
        self.context['FIGURENAMES'] = self.includes

    def _sanitise_context(self, dict_=None):
        """Removes corresponding keys to empty values from context."""
        tmp_dict = copy.deepcopy(dict_)
        for key, value in dict_.items():
            if key == 'dataset':
                continue
            if isinstance(value, (collections.OrderedDict, dict)):
                tmp_dict[key] = self._sanitise_context(value)
            elif not value:
                tmp_dict.pop(key)
        dict_ = tmp_dict
        return dict_

    def _get_figure_names(self):
        """Get the names of the figures used for the report."""
        for i in range(len(self.dataset.representations)):
            if self.dataset.representations[i].plot.description \
                    == '2D plot as scaled image.':
                self._figure_name['Figure2D'] = \
                    self.dataset.representations[i].plot.filename
            elif self.dataset.representations[i].plot.description \
                    == '1D line plot.':
                self._figure_name['Figure1D'] = \
                    self.dataset.representations[i].plot.filename
            else:
                pass


class PowerSweepAnalysisReporter(aspecd.report.LaTeXReporter):
    """Create report for power sweep analysis."""

    def __init__(self, template='', filename=''):
        super().__init__(template=template, filename=filename)
        self.dataset = cwepr.dataset.ExperimentalDataset()
        # private properties
        self._metadata = dict()
        self._tasks = dict()

    def create(self):
        """Perform all methods to generate a report."""
        #TODO:
        # nicht auf das angegebene Datenset aus den Rezept-Properties
        # verlassen. Stattdessen das Dataset, das aus Apply_To in den Kontext
        # gebaut wird, verwenden.
        # Die Datensets aus dem Kontext soweit verhauen, dass sie (leicht?)
        # in Template einbaubar sind.
        # Passendes Template bauen.

        self._prepare_metadata()
        self._get_tasks()
        # TODO: Die Figures ggf aus der Liste rausholen und in ein dict packen?
        #self._get_figure_names()
        self._create_context()
        self.context = self._sanitise_context(self.context)
        super().create()

    def _prepare_metadata(self):
        self._metadata = self.context['dataset']['metadata']
        self._metadata['parameter'] = collections.OrderedDict()
        self._collect_experimental_parameters()

    def _collect_experimental_parameters(self):
        """Collect all the metadata keys."""
        for key in self._metadata.keys():
            if key not in ['sample', 'measurement', 'parameter']:
                self._metadata['parameter'][key] = \
                    self._metadata[key]

    def _create_context(self):
        """Create a dictionary containing all data to write the report."""
        self.context['TASKS'] = self._tasks
        self.context['METADATA'] = self._metadata
        self.context['FIGURENAMES'] = self.includes

    def _get_tasks(self):
        for task, _ in enumerate(self.context['dataset']['tasks']):
            task = self.context['dataset']['tasks'][task]
            if task['kind'] in ('analysis', 'processing'):
                self._tasks[task['task'][task['kind']]['description']] = {
                    'Parameters': task['task'][task['kind']]['parameters'],
                    'Comment': task['task'][task['kind']]['comment']
                }
        fit = self.context['FITTING']
        fit_coeffs = fit.metadata.calculation.parameters['coefficients']
        further_tasks = self.context['CALCDATA'].tasks
        self._tasks[further_tasks[0]['task'].analysis.description] = {
            'Parameters': str(fit_coeffs),
            'Comment': further_tasks[0]['task'].analysis.comment
        }

    def _sanitise_context(self, dict_=None):
        tmp_dict = copy.deepcopy(dict_)
        for key, value in dict_.items():
            if key == 'dataset':
                continue
            if isinstance(value, (collections.OrderedDict, dict)):
                tmp_dict[key] = self._sanitise_context(value)
            elif not value:
                tmp_dict.pop(key)
        dict_ = tmp_dict
        return dict_


class DokuwikiCaptionsReporter(aspecd.report.Reporter):
    """Write DokuWiki Captions.

    ..todo::
        Write Documentation

    """

    def __init__(self, template='', filename=''):
        self.filename = filename
        self.language = 'de'
        self.template = template if template else self._get_template()
        super().__init__(template=self.template, filename=self.filename)
        self.dataset = cwepr.dataset.ExperimentalDataset()
        # private properties
        self._metadata = dict()
        self._figure_name = dict()

    def create(self):
        """Perform all methods to create the captions."""
        self._prepare_metadata()
        self._create_context()
        super().create()

    def _get_template(self):
        language = self.language
        module_rootpath = os.path.split(os.path.abspath(__file__))[0]
        return os.path.join(module_rootpath, 'templates', language,
                            'DokuwikiCaption.txt.jinja')

    def _prepare_metadata(self):
        self._metadata = self.context['dataset']['metadata']
        self._metadata['parameter'] = collections.OrderedDict()
        self._collect_experimental_parameters()

    def _collect_experimental_parameters(self):
        """Collect all the metadata keys."""
        for key in self._metadata.keys():
            if key not in ['sample', 'measurement', 'parameter']:
                self._metadata['parameter'][key] = \
                    self._metadata[key]

    def _create_context(self):
        self.context['METADATA'] = self._metadata


class InfofileReporter(DokuwikiCaptionsReporter):
    """Write infofile with metadata od the dataset.

    Because of it is humans who write the infofile, there may be wrong data
    especially in the spectrometer's parameters. As they usually get also
    reported in the respective data file, they are also read from it and
    stored in the dataset's metadata. A corrected infofile can be written
    herewith.

    Examples
    --------
    .. code-block:: yaml

        - kind: reporter
          type: InfofileReporter
          properties:
            parameters:
                filename: NewInfofile.yaml

    """

    def __init__(self):
        self.filename = ''
        self.language = 'en'
        self.template = self._get_template()
        super().__init__(template=self.template, filename=self.filename)

    def _get_template(self):
        language = self.language
        module_rootpath = os.path.split(os.path.abspath(__file__))[0]
        return os.path.join(module_rootpath, 'templates', language,
                            'Infofile.info.jinja')

    def _create_context(self):
        super()._create_context()
        self.context['DATASET_ID'] = \
            os.path.split(self.context['dataset']['id'])[-1]



"""Metadata.

Metadata
========

In this module, the individual metadata classes are defined which contain the
individual information about the experiment:

  * :class:`cwepr.metadata.ExperimentalDatasetMetadata`
  * :class:`cwepr.metadata.CalculatedDatasetMetadata`

What may sound like a minor detail is one key aspect of the cwepr package:
The metadata and their structure provide a unified interface for all
functionality operating on datasets. Furthermore, the metadata contained
particularly in the :class:`cwepr.metadata.ExperimentalDatasetMetadata` class
are the result of several years of practical experience. Reproducible research
is only possible if all information necessary is always recorded, and this
starts with all the metadata accompanying a measurement. Defining what kind
of metadata is important and needs to be recorded, together with metadata
formats easily writable by the experimenters *during* recording the data
requires a thorough understanding of both, the method and the setup(s) used.
For an overview of the structures of the dataset classes and their
corresponding metadata, see the :doc:`dataset structure </dataset-structure>`
section.


Module documentation
====================

"""

import aspecd.metadata
import aspecd.utils

from cwepr.exceptions import UnequalUnitsError


class DatasetMetadata(aspecd.metadata.ExperimentalDatasetMetadata):
    """Set of all metadata for a dataset object.

    Metadata as a unified structure of information coupled to the dataset are
    necessary for the understanding, analysis and processing of data,
    especially in cwepr. Too many parameters have an direct influence to the
    spectral shape of the spectrum that anything other than saving them in an
    appropriate place and accessing them automatised in the respective tasks
    is no option. Some parameters are written automatically by the
    spectrometer's software, others, depending also on the actual setup (that
    may change over time!) are omitted and it is highly recommended those
    should be noted by hand, for example in an *.info-file.*

    Attributes
    ----------
    measurement: :class:`cwepr.metadata.Measurement`
        Metadata corresponding to the measurement.

    sample : :class:`cwepr.metadata.Sample`
        Metadata corresponding to the sample.

    temperature_control : :class:`cwepr.metadata.TemperatureControl`
        Metadata corresponding to the temperature.

    experiment: :class:`cwepr.metadata.Experiment`
        Metadata corresponding to the experiment.

    spectrometer: :class:`cwepr.metadata.Spectrometer`
        Metadata corresponding to the spectrometer.

    magnetic_field: :class:`cwepr.metadata.MagneticField`
        Metadata corresponding to the magnetic field.

    bridge: :class:`cwepr.metadata.Bridge`
        Metadata corresponding to the microwave bridge.

    signal_channel: :class:`cwepr.metadata.SignalChannel`
        Metadata corresponding to signal channel.

    probehead: :class:`cwepr.metadata.Probehead`
        Metadata corresponding to the probehead used for the experiment.

    metadata_modifications : :class:`list`
        List of all modifications performed on the metadata, e.g, overrides.

    """

    def __init__(self):
        super().__init__()
        self.measurement = Measurement()
        self.experiment = Experiment()
        self.sample = Sample()
        self.spectrometer = Spectrometer()
        self.magnetic_field = MagneticField()
        self.bridge = Bridge()
        self.signal_channel = SignalChannel()
        self.probehead = Probehead()
        self.temperature_control = TemperatureControl()
        self.metadata_modifications = []


class CalculatedDatasetMetadata(aspecd.metadata.CalculatedDatasetMetadata):
    """Metadata for a calculated dataset.

    This class contains the minimal set of metadata for a dataset consisting
    of calculated data, i.e., :class:`cwepr.dataset.CalculatedDataset`.

    Metadata of actual datasets should extend this class by adding
    properties that are themselves classes inheriting from
    :class:`aspecd.metadata.Metadata`.

    Metadata can be converted to dict via
    :meth:`aspecd.utils.ToDictMixin.to_dict()`, e.g., for generating
    reports using templates and template engines.
    """


class Measurement(aspecd.metadata.Measurement):
    """Metadata corresponding to the measurement.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing properties to set.

    Attributes
    ----------
    label: :class:`str`
        Label of the sample, including the sample-number.

    """

    def __init__(self, dict_=None):
        # public properties
        self.label = ''
        super().__init__(dict_=dict_)


class Sample(aspecd.metadata.Sample):
    """Metadata corresponding to the sample .

    As this class inherits from :class:`aspecd.metadata.Sample`,
    see the documentation of the parent class for details and the full list
    of inherited attributes.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    description : :class:`str`
        Description of the measured sample.

    solvent : :class:`str`
        Name of the solvent used.

    preparation : :class:`str`
        Short details of the sample preparation.

    tube : :class:`str`
        Type and dimension of the sample tube used.

    """

    def __init__(self, dict_=None):
        # public properties
        self.description = ''
        self.solvent = ''
        self.preparation = ''
        self.tube = ''
        super().__init__(dict_=dict_)


class MagneticField(aspecd.metadata.Metadata):
    """Metadata corresponding to the magnetic field.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    start : :class:`aspecd.metadata.PhysicalQuantity`
        Lowest point of the magnetic field.

    stop : :class:`aspecd.metadata.PhysicalQuantity`
        Highest point of the magnetic field.

    sweep_width : :class:`aspecd.metadata.PhysicalQuantity`
        Width of the magnetic field sweep.

    step_width : :class:`aspecd.metadata.PhyPhysicalQuantity`
        Distance between two points (only if equidistant!).

    points : :class:`int`
        Number of points.

    field_probe_type : :class:`str`
        Type of the field probe (e.g. Hall or Teslameter)

    field_probe_model : :class:`str`
        Exact model of the field probe.

    sequence : :class:`str`
        Sequence of the experiment (e.g. up or down).

    controller : :class:`str`
        Model of the field controller.

    power_supply : :class:`str`
        Model of the power supply.

    """

    def __init__(self, dict_=None):
        super().__init__(dict_=dict_)
        self.start = aspecd.metadata.PhysicalQuantity()
        self.stop = aspecd.metadata.PhysicalQuantity()
        self.sweep_width = aspecd.metadata.PhysicalQuantity()
        self.points = int()
        self.field_probe_type = ""
        self.field_probe_model = ""
        self.sequence = ""
        self.controller = ""
        self.power_supply = ""

    def can_calculate(self):
        """Check if enough data is present to determine field values.

        Checks if enough different pieces of information are provided to
        calculate all information concerning the field sector and sweeping
        steps.

        .. note::
            Currently, the possibility of calculating the sector width from the
            width and the number of steps is not accounted for.

        Raises
        ------
        NotEnoughValuesError
            Raised when not enough different pieces of information are provided
            to determine the other variables.

        """
        sector_def = False
        sector_par = 0
        step_def = False
        step_par = 0
        if self.start.value != 0.:
            sector_def = True
            sector_par += 1
        if self.stop.value != 0.:
            sector_def = True
            sector_par += 1
        if self.sweep_width.value != 0.:
            sector_par += 1
            step_par += 1
        if self.points != 0:
            step_def = True
            step_par += 1
        if (not sector_def) or sector_par < 2:
            raise ValueError("Sector not determined.")
        if (not step_def) or step_par < 2:
            raise ValueError("Steps not determined.")
        return True

    def calculate_values(self):
        """Perform the calculation of all field values left out.

        Calculate the different values concerning the sector and sweeping steps
        of the magnetic field.

        .. note::
            Currently, the possibility of calculating the sector width from the
            with and the number of steps is not accounted for.

        Raises
        ------
        UnequalUnitsError :
            Raised, when two physical quantities shall be added or subtracted
            that have unequal units.

        """
        units_error_message = "Quantities with different units provided."
        if self.can_calculate():
            self.points = int(self.points)
            self._calc_field_width(units_error_message)
            self._calc_field_limits(units_error_message)
            self._calc_step_data(units_error_message)

    def _calc_field_width(self, units_error_message):
        if self.sweep_width.value == 0.:
            if self.stop.unit != self.start.unit:
                raise UnequalUnitsError(units_error_message)
            self.sweep_width.value = self.stop.value - self.start.value
            self.sweep_width.unit = self.stop.unit

    def _calc_field_limits(self, units_error_message):
        if self.stop.value == 0.:
            if self.sweep_width.unit != self.start.unit:
                raise UnequalUnitsError(units_error_message)
            self.stop.value = self.start.value + self.sweep_width.value
            self.stop.unit = self.start.unit
        if self.start.value == 0.:
            if self.stop.unit != self.sweep_width.unit:
                raise UnequalUnitsError(units_error_message)
            self.start.value = self.stop.value - self.sweep_width.value
            self.start.unit = self.stop.unit

    def _calc_step_data(self, units_error_message):
        if self.points == 0:
            if self.sweep_width.unit != self.step_width.unit:
                raise UnequalUnitsError(units_error_message)
            self.points = int(round((self.sweep_width.value /
                                     self.step_width.value), 0)) + 1
        if self.step_width.value == 0.:
            self.step_width.value = \
                self.sweep_width.value / (self.points - 1)
            self.step_width.unit = self.stop.unit

    def gauss_to_millitesla(self):
        """Transform magnetic field parameters from gauss to millitesla."""
        for quantity in [self.start, self.stop,
                         self.sweep_width, self.step_width]:
            quantity.value /= 10
            quantity.unit = "mT"


class Experiment(aspecd.metadata.Metadata):
    """Metadata corresponding to the experiment.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing properties to set.

    Attributes
    ----------
    type : :class:`str`

    runs : :class:`int`
        Number of recorded runs.

    variable_parameter : :class:`str`
        Parameter that is varied during the measurement, e.g. *magnetic field*

    increment : :class:`str`
        Increment the variable parameter is changed.

    harmonic : :class:`str`
        Recorded harmonic of the signal.

    """

    def __init__(self, dict_=None):
        self.type = ""
        self.runs = None
        self.variable_parameter = ""
        self.increment = None
        self.harmonic = None
        super().__init__(dict_=dict_)


class Spectrometer(aspecd.metadata.Metadata):
    """Metadata information on what type of spectrometer was used.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing properties to set.

    Attributes
    ----------
    model : :class:`str`
        Model of the spectrometer used.

    software : :class:`str`
        Name and version of the software used.

    """

    def __init__(self, dict_=None):
        self.model = ""
        self.software = ""
        super().__init__(dict_=dict_)


class Bridge(aspecd.metadata.Metadata):
    """Metadata corresponding to the microwave bridge.

    The microwave bridge contains the microwave source and parts of the
    detection system. Therefore, the crucial experimental parameters such as
    attenuation and power, microwave frequency and detection system used are
    contained as well as the description of the devices, *i.e.* the bridge
    itself, its controller, and the frequency counter, as these can be
    different interchangeable components.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    model : :class:`str`
        Model of the microwave bridge used.

    controller : :class:`str`
        Model of the bridge controller used.

    attenuation : :class:`aspecd.metadata.PhysicalQuantity`
        Attenuation of the microwave power in dB.

        Without knowing the unattenuated source power, the attenuation is a
        rather useless value, although it gets often used, particularly in
        lab jargon. Typical microwave bridges have source powers of 200 mW
        in X-Band, but newer devices sometimes deliver only 150 mW.

    power : :class:`aspecd.metadata.PhysicalQuantity`
        Output power of the microwave.

        The actual output power of the microwave used for the experiment,
        *i.e.* the source power reduced by the attenuation. Typical values
        are in the range of 20 mW to 20 µW.

    detection : :class:`str`
        Type of the detection used.

        There are two types of detection: diode and mixer. The latter
        usually allows for quadrature detection, *i.e.* detecting both, the
        absorptive and dispersive signal components.

    frequency_counter : :class:`str`
        Model of the frequency counter used.

        Depending on the setup used, this can be included in the bridge.
        Otherwise, it will often be a HP device.

    mw_frequency : :class:`aspecd.metadata.PhysicalQuantity`
        Microwave frequency.

        The actual microwave frequency used for the experiment. Usually,
        this is a scalar number. Depending on the experiment control
        software used, the microwave frequency for each transient will be
        recorded, thus allowing for analysing frequency drifts. This is
        particularly helpful in case of long-running experiments (12+ h).
        By comparing the amplitude of the frequency drift with the field
        step width, the potential impact in the signal shape can be directly
        calculated.


    q_value : :class:`int`
        Quality factor of the cavity.

        In most spectrometers, acquiring the Q-factor is not done by hand
        i.e. in Bruker spectrometers the measurement ist most commonly
        performed in tune mode with an attenuation of 33 dB, whereas at the
        Magnettech benchtop spectrometer, one has to select the box to
        measure the Q-factor.

    """

    def __init__(self, dict_=None):
        self.model = ""
        self.controller = ""
        self.attenuation = aspecd.metadata.PhysicalQuantity()
        self.power = aspecd.metadata.PhysicalQuantity()
        self.detection = ""
        self.frequency_counter = ""
        self.mw_frequency = aspecd.metadata.PhysicalQuantity()
        self.q_value = None
        super().__init__(dict_=dict_)


class SignalChannel(aspecd.metadata.Metadata):
    """Metadata information information on the signal channel employed.

    Parameters
    ----------
    dict_ : dict
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    model : :class:`str`
        Model of the signal channel.

    modulation_amplifier : :class:`str`
        Type of the modulation amplifier.

    accumulations : :class:`int`
        Number of accumulated scans.

    modulation_frequency : :class:`aspecd.metadata.PhysicalQuantity`
        Modulation frequency used.

    modulation_amplitude : :class:`aspecd.metadata.PhysicalQuantity`
        Amplitude of the modulation

    receiver_gain : :class:`aspecd.metadata.PhysicalQuantity`
        Gain of the receiver, existence depends on the spectrometer.

    conversion_time : :class:`aspecd.metadata.PhysicalQuantity`
        Conversion time (usually in ms).

    time_constant : :class:`aspecd.metadata.PhysicalQuantity`
        Time constant (usually in ms).

    phase : :class:`aspecd.metadata.PhysicalQuantity`
        Phase of the modulation amplifier.

    offset : :class:`float`
        Baseline offset

    """

    def __init__(self, dict_=None):
        self.model = ""
        self.modulation_amplifier = ""
        self.accumulations = int()
        self.modulation_frequency = aspecd.metadata.PhysicalQuantity()
        self.modulation_amplitude = aspecd.metadata.PhysicalQuantity()
        self.receiver_gain = aspecd.metadata.PhysicalQuantity()
        self.conversion_time = aspecd.metadata.PhysicalQuantity()
        self.time_constant = aspecd.metadata.PhysicalQuantity()
        self.phase = aspecd.metadata.PhysicalQuantity()
        self.offset = float()
        super().__init__(dict_=dict_)


class Probehead(aspecd.metadata.Metadata):
    """Metadata corresponding to the probehead.

    Often, resonating structures get used in EPR spectroscopy, but as this
    is not always the case, the term "probehead" is more generic.

    In all except of fully integrated benchtop spectrometers, the probehead
    can readily be exchanged. As each probehead has its own characteristics,
    it is crucially important to note at least type and model. The coupling
    (critically or overcoupled) determines the bandwidth of the resonator,
    and in all but pulsed experiments, usually, critical coupling is used.


    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing properties to set.

    Attributes
    ----------
    type : :class:`str`
        Type of the probehead used.

        There are several different types of probeheads regularly used. For
        resonators, there are, *e.g.*, dielectic and split-ring resonators,
        cylindrical and rectangular cavities. More special would be
        Fabry-Perot and stripline resonators. Sometimes, even resonator-free
        designs are used as probeheads.


    model : :class:`str`
        Model of the probehead used.

        Commercial probeheads come with a distinct model that goes in here.
        In all other cases, use a short, memorisable, and unique name.


    coupling : :class:`str`
        Type of coupling. In cwepr it is sually critically coupled.

    """

    def __init__(self, dict_=None):
        self.type = ""
        self.model = ""
        self.coupling = ""
        super().__init__(dict_=dict_)


class TemperatureControl(aspecd.metadata.TemperatureControl):
    """Metadata corresponding to the temperature control.

    As this class inherits from :class:`aspecd.metadata.TemperatureControl`,
    see the documentation of the parent class for details and the full list
    of inherited attributes.


    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing properties to set.

    Attributes
    ----------
    cryostat : :class:`str`
        Model of the cryostat used.

    cryogen : :class:`str`
        Cryogen used.

        Typically, this is either N2 (for temperatures down to 80K) or He
        (for temperatures down to 4 K)

    """

    def __init__(self, dict_=None):
        # public properties
        self.cryostat = ''
        self.cryogen = ''
        super().__init__(dict_=dict_)



"""
Subpackage for IO.

Generally, for each file format, the importer resides in a separate module.
The import statements below should *only* import the respective classes.

Currently, in addition to modules for the individual data file formats,
there are a series of more general modules:



* factory

  Factory classes, currently the DatasetImporterFactory

* exporter

  Exporters, currently only an ASCII exporter


"""

from .factory import DatasetImporterFactory
from .magnettech import MagnettechXmlImporter
from .txt_file import TxtImporter
from .bes3t import BES3TImporter
from .exporter import ASCIIExporter, MetadataExporter



"""Datasets: units containing data and metadata.

The dataset is one key concept of the ASpecD framework and hence the cwepr
package derived from it, consisting of the data as well as the corresponding
metadata. Storing metadata in a structured way is a prerequisite for a
semantic understanding within the routines. Furthermore, a history of every
processing, analysis and annotation step is recorded as well, aiming at a
maximum of reproducibility. This is part of how the ASpecD framework and
therefore the cwepr package tries to support good scientific practice.

Therefore, each processing and analysis step of data should always be
performed using the respective methods of a dataset, at least as long as it
can be performed on a single dataset.

Datasets
========

Generally, there are two types of datasets: Those containing experimental
data and those containing calculated data. Therefore, two corresponding
subclasses exist:

  * :class:`cwepr.dataset.ExperimentalDataset`
  * :class:`cwepr.dataset.CalculatedDataset`


Dataset factory
===============

Particularly in case of recipe-driven data analysis (c.f. :mod:`aspecd.tasks`),
there is a need to automatically retrieve datasets using nothing more than a
source string that can be, e.g., a path or LOI. This is where the
DatasetFactory comes in. This is a factory in the sense of the factory
pattern described by the "Gang of Four" in their seminal work, "Design
Patterns" (Gamma et al., 1995):

  * :class:`cwepr.dataset.DatasetFactory`


Module documentation
====================
"""

import aspecd.dataset
import aspecd.utils

import cwepr.io.factory
import cwepr.metadata


class ExperimentalDataset(aspecd.dataset.ExperimentalDataset):
    """Set of data uniting all relevant information.

    The unity of numerical and metadata is indispensable for the
    reproducibility of data and is possible by saving all information available
    for one set of measurement data in a single instance of this class.
    """

    def __init__(self):
        super().__init__()
        self.metadata = cwepr.metadata.DatasetMetadata()


class CalculatedDataset(aspecd.dataset.CalculatedDataset):
    """Entity consisting of calculated data and metadata.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.dataset.CalculatedDataset`
    class for details.

    """


class DatasetFactory(aspecd.dataset.DatasetFactory):
    """
    Factory for creating dataset objects based on the source provided.

    Particularly in case of recipe-driven data analysis (c.f.
    :mod:`aspecd.tasks`),
    there is a need to automatically retrieve datasets using nothing more
    than a source string that can be, e.g., a path or LOI.

    The DatasetFactory operates in conjunction with a
    :class:`cwepr.io.factory.DatasetImporterFactory` to import the actual
    dataset. See the respective class documentation for more details.


    Attributes
    ----------
    importer_factory : :class:`cwepr.io.factory.DatasetImporterFactory`
        ImporterFactory instance used for importing datasets

    """

    def __init__(self):
        super().__init__()
        self.importer_factory = cwepr.io.factory.DatasetImporterFactory()

    @staticmethod
    def _create_dataset(source=''):
        """Return cwepr dataset.

        Parameters
        ----------
        source : :class:`str`
            string describing the source of the dataset

            May be a filename or path, a URL/URI, a LOI, or similar

        Returns
        -------
        dataset : :class:`cwepr.dataset.ExperimentalDataset`
            Dataset object for cwepr package

        """
        return cwepr.dataset.ExperimentalDataset()



import os
import shutil
import tempfile
import unittest

import numpy as np

import cwepr.dataset
import cwepr.exceptions
import cwepr.io.magnettech
import cwepr.processing

ROOTPATH = os.path.split(os.path.abspath(__file__))[0]


class TestMagnettechXmlImporter(unittest.TestCase):
    def setUp(self):
        source = os.path.join(ROOTPATH, 'testdata/test-magnettech')
        self.importer = cwepr.io.magnettech.MagnettechXmlImporter(source=source)

    def test_axis_dimensions_equals_one(self):
        converter = cwepr.io.magnettech.MagnettechXmlImporter()
        testdata = 'CZzAKavudEA=5HabpLDudEA='
        self.assertEqual(1, converter._convert_base64string_to_np_array(
                             testdata).ndim)

    def test_specific_fields_are_filled(self):
        dataset = cwepr.dataset.ExperimentalDataset()
        dataset.import_from(self.importer)

        # arbitrary attributes that must have been set
        teststr = ['temperature_control.temperature.value',
                   'magnetic_field.start.unit',
                   'bridge.mw_frequency.value']
        for string_ in teststr:
            metadata_object = dataset.metadata
            for e in string_.split('.'):
                metadata_object = getattr(metadata_object, e)
            self.assertTrue(metadata_object)

    def test_q_value_is_float(self):
        dataset = cwepr.dataset.ExperimentalDataset()
        dataset.import_from(self.importer)
        q_value = dataset.metadata.bridge.q_value
        self.assertEqual(float, type(q_value))

    def test_import_with_no_file_raises(self):
        importer = cwepr.io.magnettech.MagnettechXmlImporter()
        dataset = cwepr.dataset.ExperimentalDataset()
        with self.assertRaises(cwepr.exceptions.MissingPathError):
            dataset.import_from(importer)

    def test_import_with_not_existing_file_raises(self):
        source = 'foo.xml'
        importer = cwepr.io.magnettech.MagnettechXmlImporter(source=source)
        dataset = cwepr.dataset.ExperimentalDataset()
        with self.assertRaises(FileNotFoundError):
            dataset.import_from(importer)

    def test_import_with_no_infofile_continues(self):
        source = os.path.join(ROOTPATH, 'testdata/test-magnettech.xml')
        with tempfile.TemporaryDirectory() as testdir:
            new_source = os.path.join(testdir, 'test-wo-infofile')
            shutil.copyfile(source, new_source + '.xml')
            dataset = cwepr.dataset.ExperimentalDataset()
            importer = cwepr.io.magnettech.MagnettechXmlImporter(
                source=new_source)
            dataset.import_from(importer)

    def test_with_file_extension(self):
        source = os.path.join(ROOTPATH, 'testdata/test-magnettech.xml')
        dataset = cwepr.dataset.ExperimentalDataset()
        importer = cwepr.io.magnettech.MagnettechXmlImporter(
            source=source)
        dataset.import_from(importer)


class TestGoniometerSweepImporter(unittest.TestCase):
    def setUp(self):
        source = os.path.join(ROOTPATH, 'testdata/magnettech-goniometer/')
        self.goniometer_importer = \
            cwepr.io.magnettech.GoniometerSweepImporter(source=source)

    def instantiate_class(self):
        pass

    def test_has_import_method(self):
        self.assertTrue(hasattr(self.goniometer_importer, '_import'))
        self.assertTrue(callable(self.goniometer_importer._import))

    def test_source_path_doesnt_exist_raises(self):
        source = 'foo/'
        importer = cwepr.io.magnettech.GoniometerSweepImporter(source=source)
        dataset = cwepr.dataset.ExperimentalDataset()
        with self.assertRaises(FileNotFoundError):
            dataset.import_from(importer)

    def test_sort_filenames_returns_sorted_list(self):
        self.goniometer_importer._get_filenames()
        self.goniometer_importer._sort_filenames()
        sorted_list = self.goniometer_importer.filenames
        nums = []
        for filename in sorted_list:
            num = filename.split('gon_')[1]
            nums.append(num.split('dg')[0])
        for x in range(len(nums)-1):
            self.assertGreater(int(nums[x+1])-int(nums[x]), 0)

    def test_has_import_all_data_to_list_method(self):
        self.assertTrue(hasattr(self.goniometer_importer,
                                '_import_all_spectra_to_list'))
        self.assertTrue(callable(
            self.goniometer_importer._import_all_spectra_to_list))

    def test_angles_smaller_than_360_deg(self):
        dataset = cwepr.dataset.ExperimentalDataset()
        dataset.import_from(self.goniometer_importer)
        self.assertTrue(all([x < 359 for x in
                             self.goniometer_importer._angles]))

    def test_import_data_fills_dataset(self):
        dataset = cwepr.dataset.ExperimentalDataset()
        dataset.import_from(self.goniometer_importer)
        self.assertNotEqual(0, dataset.data.data.size)

    def test_data_and_filenames_have_same_lengths(self):
        # Check whether all data has been imported correctly and was moved
        # entirely to the final dataset.
        dataset = cwepr.dataset.ExperimentalDataset()
        dataset.import_from(self.goniometer_importer)
        self.assertEqual(len(self.goniometer_importer.filenames),
                         self.goniometer_importer.dataset.data.data.shape[1])

    def test_all_datasets_have_same_frequency(self):
        dataset = cwepr.dataset.ExperimentalDataset()
        dataset.import_from(self.goniometer_importer)
        frequencies = np.array([])
        for set_ in self.goniometer_importer._data:
            frequencies = np.append(frequencies,
                                    set_.metadata.bridge.mw_frequency.value)
        self.assertAlmostEqual(max(frequencies), min(frequencies))


    def test_goniometer_import(self):
        source = os.path.join(ROOTPATH, 'testdata/magnettech-goniometer/')
        dataset = cwepr.dataset.ExperimentalDataset()
        importer = cwepr.io.magnettech.GoniometerSweepImporter(
            source=source)
        dataset.import_from(importer)



"""Functionalities to simply import a txt file containing data."""
import numpy as np

import aspecd.io
import cwepr.processing
import cwepr.dataset


class TxtImporter(aspecd.io.DatasetImporter):
    """Simple importer for txt files containing data.

    Sometimes, data come from sources as a txt file only. So far, the format
    is hardcoded and should contain two columns separated by a tabulator.

    """

    def __init__(self, source=''):
        super().__init__(source=source)
        # public properties
        self.extension = '.txt'

    def _import(self):
        self._get_data()
        self._create_metadata()

    def _get_data(self):
        self.source = self.source + self.extension
        raw_data = np.loadtxt(self.source, delimiter='\t')
        self.dataset.data.data = raw_data[:, 1]
        self.dataset.data.axes[0].values = raw_data[:, 0]

    def _create_metadata(self):
        self.dataset.data.axes[0].unit = 'mT'
        self.dataset.data.axes[1].quantity = 'intensity'


class CsvImporter(aspecd.io.DatasetImporter):
    """Importer for simple csv imports with different delimiters."""

    def __init__(self, source=''):
        super().__init__(source=source)
        # public properties
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.extension = '.csv'

    def _import(self):
        self._read_data()

    def _read_data(self):
        with open(self.source + self.extension) as csv_file:
            # noinspection PyTypeChecker
            raw_data = np.loadtxt(csv_file, delimiter=';', skiprows=3)
            self.dataset.data.data = raw_data[:, 1]
            self.dataset.data.axes[0].values = raw_data[:, 0]

    def _create_metadata(self):
        self.dataset.data.axes[0].unit = 'mT'
        self.dataset.data.axes[1].quantity = 'intensity'



"""Exporters for datasets.

.. note::
    not tested.

"""
import collections
import datetime

import numpy as np
import aspecd.io
import aspecd.utils


class ASCIIExporter(aspecd.io.DatasetExporter):
    """Export a dataset in ASCII format.

    Exports the complete dataset to an ASCII file. At the same time, the
    respective metadata is exported into a YAML file using the functionality
    provided by aspecd. Metadata residing there as an object of the class
    :class:`numpy.ndarray` are converted into lists.
    """

    def __init__(self):
        super().__init__()

    def _export(self):
        """Export the dataset's numeric data and metadata."""
        file_name_data = self.target + ".txt"
        file_name_meta = self.target + ".yaml"
        np.savetxt(file_name_data, self.dataset.data.data, delimiter=",")
        metadata_writer = aspecd.utils.Yaml()
        metadata = self._get_and_prepare_metadata()
        metadata_writer.dict = metadata
        metadata_writer.write_to(filename=file_name_meta)

    def _get_and_prepare_metadata(self):
        """Prepare the dataset's metadata to be imported.

        Transforms the metadata to a dict and subsequently transforms all
        instances of :class:`numpy.ndarray` by transforming them into lists.

        Returns
        -------
        metadata_prepared: :class:`dict`
            transformed metadata

        """
        metadata = self.dataset.metadata.to_dict()
        metadata_prepared = self._ndarrays_to_list_recursively(metadata)
        return metadata_prepared

    def _ndarrays_to_list_recursively(self, dictionary):
        """Transforms instances of :class:`numpy.array` from a given dict.

        Numpy arrays are not easily convertible into a yaml-file and are
        therefore transformed to lists.
        This is a cascaded method and thus also works on nested dicts.

        Parameters
        ----------
        dictionary: :class:`dict`
            Dictionary to relieve of arrays.

        Returns
        -------
        dictionary: :class:`dict`
            Dictionary relieved of arrays.

        """
        for key, value in dictionary.items():
            if isinstance(value, dict):
                dictionary[key] = self._ndarrays_to_list_recursively(value)
            if isinstance(value, np.ndarray):
                dictionary[key].to_list()
        return dictionary


class MetadataExporter(aspecd.io.DatasetExporter):
    """Export metadata to yaml file with default name."""

    def __init__(self):
        super().__init__()
        self.metadata_dict = collections.OrderedDict()
        self.filename = ''

    def _export(self):
        if not self.filename:
            timestamp = datetime.datetime.now().strftime('%Y%m%dT%H%M%S')
            self.filename = 'metadata-' + timestamp + '.yaml'
        if not self.filename.endswith(('.yaml', '.yml')):
            self.filename = self.filename + '.yaml'
        self._write_metadata()

    def _write_metadata(self):
        self.metadata_dict = self.dataset.metadata.to_dict()
        self.metadata_dict = \
            self._remove_empty_items_recursively(dict_=self.metadata_dict)
        yaml_file = aspecd.utils.Yaml()
        yaml_file.dict = self.metadata_dict
        yaml_file.write_to(self.filename)

    def _remove_empty_items_recursively(self, dict_=None):
        tmp_dict = collections.OrderedDict()
        for key, value in dict_.items():
            if isinstance(value, dict):
                dict_[key] = \
                    self._remove_empty_items_recursively(value)
            # if magnettech has not measured the q-value it is -1
            if key == 'q_value' and value == -1:
                continue
            if dict_[key]:
                tmp_dict[key] = dict_[key]
        return tmp_dict



import os
import setuptools

with open(os.path.join(os.path.dirname(__file__), 'VERSION')) as version_file:
    version = version_file.read().strip()

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
    readme = f.read()


setuptools.setup(
    name='cwepr',
    version=version,
    description='Package for handling cw-EPR data.',
    long_description=readme,
    long_description_content_type='text/x-rst',
    author='Mirjam Schröder, Pascal Kirchner, Till Biskup',
    author_email='till@till-biskup.de',
    url='https://www.cwepr.de/',
    project_urls={
        "Documentation": "https://docs.cwepr.de/",
        "Source": "https://github.com/tillbiskup/cwepr",
    },
    packages=setuptools.find_packages(exclude=('tests', 'docs')),
    include_package_data=True,
    license='BSD',
    keywords=[
        'EPR spectroscopy',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Development Status :: 4 - Beta",
        "Topic :: Scientific/Engineering",
    ],
    install_requires=[
        'aspecd>=0.2.1',
        'numpy',
        'scipy',
        'matplotlib',
    ],
    python_requires='>=3.5',
)



"""Module containing the analysis steps of the cwEPR package.

.. sidebar::
    processing *vs.* analysis

    For more details on the difference between processing and analysis,
    see the `ASpecD documentation <https://docs.aspecd.de/>`_.


An analysis step, in contrast to a processing step, does not simply modify a
given dataset, but extracts some characteristic of this dataset that is
contained in its ``results`` property. This can be as simple as a number,
*e.g.*, in case of the signal-to-noise figure of a data trace, but as
complicated as a (calculated) dataset on its own containing some measure as
function of some parameter, such as the microwave frequency for each of a
series of recordings, allowing to visualise drifts that may or may not
impact data analysis.

.. note::
    This module may be split into different modules and thus converted into
    a subpackage at some point in the future, depending on the number of
    classes for analysis steps to come (but rather likely to happen).

.. todo::
    Make methods dealing with both, 1D and 2D datasets or raising the
    respective errors.


Note to developers
==================

Processing steps can be based on analysis steps, but not inverse! Otherwise,
we get cyclic dependencies what should obviously be avoided in order to keep
code working.
"""

import copy
import math
import numpy as np
import scipy.constants

import aspecd.analysis
import aspecd.dataset

from cwepr.exceptions import MissingInformationError


class FieldCorrectionValue(aspecd.analysis.SingleAnalysisStep):
    """Determine correction value for a field correction.

    As g standard reference, it can be chosen from two substances, LiLiF and
    DPPH using the parameter "standard".

    .. todo::
        Check for units and make them consistent for both points.

    References for the constants:

    g value of Li:LiF:

        g(LiLiF) = 2.002293 +- 0.000002

    Reference: Rev. Sci. Instrum. 1989, 60, 2949-2952.

    g value of DPPH:

        g(DPPH) = 2.0036 ± 0.0002

    Reference: Appl. Magn. Reson. 10, 339-350 (1996)

    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        standard : :class:`str`
            Field standard that should be applied.
            Valid values are "LiLiF" and "DPPH"

            Default: LiLiF

        mw_frequency : :class:`float`
            Microwave frequency to be corrected for. In general, is taken
            from the dataset, what is recommended, but can also be given.

    Raises
    ------
    MissingInformationError
        Raised if no microwave frequency is given neither in the dataset nor
        as parameter.

    Examples
    --------
    .. code-block:: yaml

        - kind: singleanalysis
          type: FieldCorrectionValue
          properties:
            parameters:
                mw_frequency: 9.5
                standard: LiLiF
          result: deltaB

    """

    G_LILIF = 2.002293
    G_DPPH = 2.0036

    def __init__(self):
        super().__init__()
        self.parameters['mw_frequency'] = None
        self.parameters['standard'] = 'LiLiF'
        self.description = "Determination of a field correction value"
        self._g_correct = None

    def _perform_task(self):
        """Wrapper around field correction value determination method."""
        self._get_mw_frequency()
        self._get_sample_g_value()
        self.result = self._get_field_correction_value()

    def _get_mw_frequency(self):
        if not self.parameters['mw_frequency'] and not \
                self.dataset.metadata.bridge.mw_frequency.value:
            raise MissingInformationError(message='No microwave frequency '
                                                  'provided, aborting.')
        if not self.parameters['mw_frequency']:
            self.parameters['mw_frequency'] = \
                self.dataset.metadata.bridge.mw_frequency.value

    def _get_field_correction_value(self):
        """Calculates a field correction value.

        Finds the zero-crossing of a (derivative) spectrum of a field
        standard by using the difference between minimum and maximum part of the
        signal. This value is then subtracted from the the expected field
        value for the MW frequency provided.

        Returns
        -------
        delta_b0: :class:`float`
            Field correction value

        """
        i_max = np.argmax(self.dataset.data.data)
        i_min = np.argmin(self.dataset.data.data)
        zero_crossing_exp = (self.dataset.data.axes[0].values[i_min] +
                             self.dataset.data.axes[0].values[i_max]) / 2
        calculated_field = \
            scipy.constants.value('Planck constant') \
            * self.parameters['mw_frequency'] * 1e9 * 1e3 / \
            (self._g_correct * scipy.constants.value('Bohr magneton'))
        delta_b0 = calculated_field - zero_crossing_exp
        return delta_b0

    def _get_sample_g_value(self):
        if 'standard' in self.parameters.keys():
            if self.parameters['standard'].lower() == 'dpph':
                self._g_correct = self.G_DPPH
            elif self.parameters['standard'].lower() == 'lilif':
                self._g_correct = self.G_LILIF
            else:
                print('This g-standard sample is not listed, LiLiF is chosen '
                      'automatically.')


class LinewidthPeakToPeak(aspecd.analysis.SingleAnalysisStep):
    """Peak to peak linewidth in derivative spectrum.

    The linewidth is given in a dirst derivative spectrum as difference
    between the two extreme points. However, this is valid only for simple
    spectra with just one line or signal. This analysis step simply takes the
    difference on the magnetic field axis which is then stored in the result.
    The task can be used as following:

    .. code-block:: yaml

      - kind: singleanalysis
        type: LinewidthPeakToPeak
        result: linewidth

    """

    def __init__(self):
        super().__init__()
        self.description = "Determine peak-to-peak linewidth"

    def _perform_task(self):
        self.result = self.get_peak_to_peak_linewidth()

    @staticmethod
    def applicable(dataset):  # noqa: D102
        return isinstance(dataset.data.data.size, int)

    def get_peak_to_peak_linewidth(self):
        """Calculates the peak-to-peak linewidth.

        This is done by determining the distance between the maximum and the
        minimum in the derivative spectrum which should yield acceptable
        results in a symmetrical signal. Limitation: Spectrum contains only
        one line.

        Returns
        -------
        linewidth: :class:`float`
            peak to peak linewidth

        """
        index_max = np.argmax(self.dataset.data.data)
        index_min = np.argmin(self.dataset.data.data)
        linewidth = abs(self.dataset.data.axes[0].values[index_min] -
                        self.dataset.data.axes[0].values[index_max])
        return linewidth


class LinewidthFWHM(aspecd.analysis.SingleAnalysisStep):
    """Full linewidth at half maximum (FWHM).

    In EPR, this linewidth can be applied to integrated cwEPR spectra or
    spectra recorded in absorptive mode.
    The calculation is done by subtracting maximum/2, getting the absolute
    value and then determining the minima. The distance between these points
    corresponds to the FWHM linewidth.

    Examples
    --------
    Usage is quite simple as it requires no additional parameters:

    .. code-block:: yaml

        - kind: singleanalysis
          type: LinewidthFWHM
          result: linewidth


    """

    def __init__(self):
        super().__init__()
        self.description = "Determine linewidth (full width at half max; FWHM)"

    def _perform_task(self):
        self.result = self._get_fwhm_linewidth()

    def _get_fwhm_linewidth(self):
        """Calculates the line width (full width at half maximum, FWHM).

        Returns
        -------
        linewidth: :class:`float`
            line width as determined

        """
        index_max = np.argmax(self.dataset.data.data)
        spectral_data = copy.deepcopy(self.dataset.data.data)

        spectral_data -= max(spectral_data) / 2
        spectral_data = abs(spectral_data)

        left_zero_cross_index = np.argmin(spectral_data[:index_max])
        right_zero_cross_index = np.argmin(spectral_data[index_max:])
        b_field_left = self.dataset.data.axes[0].values[left_zero_cross_index]
        b_field_right = self.dataset.data.axes[0].values[right_zero_cross_index]
        return b_field_right - b_field_left


class SignalToNoiseRatio(aspecd.analysis.SingleAnalysisStep):
    """Get a spectrum's signal to noise ratio.

    This is done by comparing the absolute maximum of the spectrum to the
    maximum of the edge part of the spectrum (i.e. a part which is considered
    to not contain any signal.

    Attributes
    ----------
    parameters['percentage']: :class:`int`
        percentage of the spectrum to be considered edge part on any side
        (i.e. 10 % means 10 % on each side).

        Default: 10 %

    Examples
    --------
    The analysis can be applied either with or without the additional
    parameter "percentage":

    .. code-block:: yaml

        - kind: singleanalysis
          type: SignalToNoiseRatio
          properties:
            parameters:
                percentage: 7
          result: SNR

    """

    def __init__(self):
        super().__init__()
        self.parameters["percentage"] = 10
        self.description = "Determine signal to noise ratio."

    def _perform_task(self):
        """Determine signal to noise ratio.

        Call method to get the amplitude of the noise, compare it to the
        absolute amplitude and set a result.
        """
        signal = self.dataset.data.data
        noise = self._get_noise()

        noise_amplitude = self._get_amplitude(noise)
        signal_amplitude = self._get_amplitude(signal) - noise_amplitude

        self.result = signal_amplitude / noise_amplitude

    @staticmethod
    def _get_amplitude(data=None):
        return max(data) - min(data)

    def _get_noise(self):
        number_of_points = math.ceil(len(self.dataset.data.data) *
                                     self.parameters["percentage"] / 100.0)
        noise_data = np.append(self.dataset.data.data[:number_of_points],
                               self.dataset.data.data[-number_of_points:])
        return noise_data


class Amplitude(aspecd.analysis.SingleAnalysisStep):
    """Determine amplitude of dataset.

    Depending of the dimension of the dataset, the returned value is either
    a scalar (1D dataset) or a vector (2D dataset) containing the amplitude of
    each row respectively.

    Returns
    -------
    result
        amplitude(s) row-wise, thus scalar or vector.

    Examples
    --------
    The analysis is again called without additional parameters:

    .. code-block:: yaml

        - kind: singleanalysis
          type: Amplitude
          result: amplitude

    """

    def __init__(self):
        super().__init__()
        self.description = "Get amplitude of (derivative) spectrum."

    def _perform_task(self):
        if len(self.dataset.data.axes) == 2:
            self.result = max(self.dataset.data.data) - min(
                self.dataset.data.data)
        else:
            self.result = np.amax(self.dataset.data.data, axis=0) - np.amin(
                self.dataset.data.data, axis=0)


class AmplitudeVsPower(aspecd.analysis.SingleAnalysisStep):
    """Return a calculated dataset to further analyse a power-sweep experiment.

    The analysis of a power sweep results in a plot of the peak amplitude
    vs. the square root of the microwave power of the bridge. Both values
    are determined in this step and put together in a calculated dataset
    that can be used subsequently to perform the linear fit.

    Returns
    -------
    result: :class:`aspecd.dataset.CalculatedDataset`
       Calculated Dataset where the data is the amplitude and the axis
       values is the root of the mw-power (in ascending order).

    Examples
    --------
    For analysing a power sweep, extracting the amplitude and taking the root of
    the microwave power is the first step to success (see
    :ref:`power_sweep_analysis` for further details) and can be done as
    follows without additional parameters:

    .. code-block:: yaml

        - kind: singleanalysis
          type: AmplitudeVsPower
          result: calc_dataset

    """

    def __init__(self):
        super().__init__()
        self.description = "Return calculated dataset for power sweep analysis."
        self.result = aspecd.dataset.CalculatedDataset()
        # private properties
        self._analysis = Amplitude()
        self._roots_of_mw_power = np.ndarray([])

    @staticmethod
    def applicable(dataset):  # noqa: D102
        return len(dataset.data.axes) > 2

    def _perform_task(self):
        self._calculate_data_and_axis()
        self._check_mw_axis()
        self._assign_data_to_result()
        self._assign_units_to_result()

    def _calculate_data_and_axis(self):
        self._roots_of_mw_power = np.sqrt(self.dataset.data.axes[1].values)
        amplitude = Amplitude()
        self._analysis = self.dataset.analyse(amplitude)

    def _check_mw_axis(self):
        """Check for ascending order.

        If not, revert both, root of mw-power and amplitude.
        """
        if self._roots_of_mw_power[-1] - self._roots_of_mw_power[0] < 0:
            self._roots_of_mw_power = self._roots_of_mw_power[::-1]
            self._analysis.result = self._analysis.result[::-1]

    def _assign_data_to_result(self):
        self.result.data.data = self._analysis.result
        self.result.data.axes[0].values = self._roots_of_mw_power

    def _assign_units_to_result(self):
        if self.dataset.data.axes[1].unit == 'mW':
            self.result.data.axes[0].unit = 'sqrt(mW)'
        elif self.dataset.data.axes[1].unit == 'W':
            self.result.data.axes[0].unit = 'sqrt(W)'
        self.result.data.axes[0].quantity = 'root of mw power'


class PolynomialFitOnData(aspecd.analysis.SingleAnalysisStep):
    """Perform polynomial fit on data and return its parameters or a dataset.

    Developed tests first.

    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        points
            first n points that should taken into account

            Default: 3

        order
            order of the fit.

            Default: 1

        return_type : :class: `str`
            Choose to returning the coefficients of the fit or a calculated
            dataset containing the curve to plot.

            Default: coefficients.

            Valid values: 'coefficients', 'dataset'

        add_origin : :class: `bool`
            Adds the point (0,0) to the data and axes, but  does not guarantee
            the fit really passes the origin.

            Default: False.

    Examples
    --------
    Some Parameters can be chosen here, depending on the purpose and the
    following analysis and processing steps.

    .. code-block:: yaml

        - kind: singleanalysis
          type: PolynomialFitOnData
          properties:
            parameters:
                points: 5
                order: 2
                return_type: dataset
                add_origin: True
          result: fit

    """

    def __init__(self):
        super().__init__()
        self.description = "Perform polynomial fit and return parameters."
        self.result = None
        self.parameters['points'] = 3
        self.parameters['order'] = 1
        self.parameters['return_type'] = 'coefficients'
        self.parameters['add_origin'] = False
        self.parameters['coefficients'] = []
        # private properties
        self._curve = None

    def _perform_task(self):
        if self.parameters['add_origin']:
            self._add_origin()
        self._get_coefficients()
        self._get_curve()
        self._assign_result()

    def _get_coefficients(self):
        x_data_to_process = \
            self.dataset.data.axes[0].values[:self.parameters['points']]
        y_data_to_process = self.dataset.data.data[:self.parameters['points']]
        self.parameters['coefficients'] = \
            np.polyfit(x_data_to_process, y_data_to_process, self.parameters[
                'order'])

    def _get_curve(self):
        self._curve = self.create_dataset()
        self._curve.data.axes[0] = self.dataset.data.axes[0]
        self._curve.data.data = np.polyval(self.parameters['coefficients'],
                                           self.dataset.data.axes[0].values)
        self._curve.data.axes[0].values = self.dataset.data.axes[0].values

    def _assign_result(self):
        if self.parameters['return_type'].lower() == 'dataset':
            self.result = self._curve
        else:
            self.result = self.parameters['coefficients']

    def _add_origin(self):
        self.dataset.data.axes[0].values = \
            np.insert(self.dataset.data.axes[0].values, 0, 0)
        self.dataset.data.data = np.insert(self.dataset.data.data, 0, 0)


class PtpVsModAmp(aspecd.analysis.SingleAnalysisStep):
    """Create calculated dataset for modulation sweep analysis.

    For a modulation sweep analysis, the first step is to get the peak to
    peak amplitude and correlate it to the modulation amplitude,
    see :ref:`modulation_sweep_analysis` for further details.

    Examples
    --------
    The usage is also quite simple without additional parameters necessary:

    .. code-block:: yaml

        - kind: singleanalysis
          type: PtpVsModAmp
          result: calc_dataset

    """

    def __init__(self):
        super().__init__()
        self.description = 'Create dataset with ptp-linewidth vs modulation ' \
                           'Amplitude.'
        self.new_dataset = aspecd.dataset.CalculatedDataset()
        self.linewidths = np.ndarray([0])

    def _perform_task(self):
        self._get_linewidths()
        self._fill_dataset()
        self.result = self.new_dataset

    @staticmethod
    def applicable(dataset):  # noqa: D102
        return len(dataset.data.axes) > 2

    def _get_linewidths(self):
        for line in self.dataset.data.data:
            index_max = np.argmax(line)
            index_min = np.argmin(line)
            linewidth = abs(self.dataset.data.axes[1].values[index_min] -
                            self.dataset.data.axes[1].values[index_max])
            self.linewidths = np.append(self.linewidths, linewidth)

    def _fill_dataset(self):
        self.new_dataset.data.data = self.linewidths
        self.new_dataset.data.axes[0] = self.dataset.data.axes[1]
        self.new_dataset.data.axes[1].unit = self.dataset.data.axes[0].unit
        self.new_dataset.data.axes[1].quantity = 'peak to peak linewidth'


class AreaUnderCurve(aspecd.analysis.SingleAnalysisStep):
    """Make definite integration, i.e. calculate the area under the curve.

    Takes no other parameters and returns a single number. However, this step
    does not (yet) account for baselines or noise thus that the value
    obtained here is to be taken with care.

    Examples
    --------
    .. code-block:: yaml

        - kind: singleanalysis
          type: AreaUnderCurve
          result: area

    """

    def __init__(self):
        super().__init__()
        self.description = "Definite integration / area und the curve"

    def _perform_task(self):
        """Perform the actual integration.

        The x values from the dataset are used.
        """
        x_values = self.dataset.data.axes[0].values
        y_values = self.dataset.data.data

        self.result = np.trapz(y_values, x_values)



"""Import magnettech 1D and 2D datasets from its xml file(s)."""
import base64
import glob
import os
import struct
import xml.etree.ElementTree as et
import numpy as np

import aspecd.annotation
import aspecd.infofile
import aspecd.io
import aspecd.metadata

import cwepr.dataset
import cwepr.metadata
import cwepr.processing
import cwepr.exceptions


class MagnettechXmlImporter(aspecd.io.DatasetImporter):
    """Import cw-EPR raw data from the Magnettech benchtop spectrometer.

    Magnettech provides a XML-file with the results. Specialities of this
    format are existing and will be briefly explained: The data is coded in
    hex-numbers, and the y axis consists of 10 times more points than the
    y-axis. Therefore, an interpolation is needed to expand the axis to the
    necessary amount of points.


    Attributes
    ----------
    root: :class:`str`
        path of the root directory

    .. todo::
        Dokumentation der Attribute, Sortieren der Funktionen.

    """

    def __init__(self, source=''):
        if source.endswith('.xml'):
            source = source[:-4]
        super().__init__(source=source)
        # public properties
        self.load_infofile = True
        self.root = None
        self.xml_metadata = dict()
        self.full_filename = ''
        # private properties
        self._infofile = aspecd.infofile.Infofile()
        self._bfrom = float()
        self._bto = float()
        self._xvalues = None
        self._yvalues = None
        self._infofile = aspecd.infofile.Infofile()

    def _import(self):
        self._get_full_filename()
        self._get_xml_root_element()

        self._get_raw_data()
        self._create_x_axis()
        self._extract_metadata_from_xml()  # Is needed for cutting data

        self._cut_data()
        self._hand_data_to_dataset()

        # First import metadata from infofile, then override hand-written
        # information by metadata from xml-file. The order matters.
        if self.load_infofile and self._infofile_exists():
            self._load_infofile()
            self._map_infofile()
        self._map_metadata_from_xml()

    def _get_full_filename(self):
        if self.source:
            self.full_filename = self.source + '.xml'

    def _get_xml_root_element(self):
        """Get the root object/name of the xml document."""
        if not self.source:
            raise cwepr.exceptions.MissingPathError('No path provided')
        if not os.path.exists(self.full_filename):
            raise FileNotFoundError('XML file not found.')
        self.root = et.parse(self.full_filename).getroot()

    def _get_raw_data(self):
        self._xvalues = \
            self._convert_base64string_to_np_array(self.root[0][0][1][0].text)
        self._yvalues = \
            self._convert_base64string_to_np_array(self.root[0][0][1][-1].text)

    @staticmethod
    def _convert_base64string_to_np_array(string):
        # Split string at "=" and add the delimiter afterwards again
        tmpdata = [x + "=" for x in string.split("=") if x]
        # Decode and unpack list of strings
        data = [struct.unpack('d', base64.b64decode(x)) for x in tmpdata]
        data = [i[0] for i in data]
        return np.asarray(data)

    def _create_x_axis(self):
        b_field_x_offset = float(self.root[0][0][1][0].attrib['XOffset'])
        b_field_x_slope = float(self.root[0][0][1][0].attrib['XSlope'])
        mw_abs_x_offset = float(self.root[0][0][1][-1].attrib['XOffset'])
        mw_abs_x_slope = float(self.root[0][0][1][-1].attrib['XSlope'])

        mw_x = mw_abs_x_offset + \
            np.linspace(0, len(self._yvalues) - 1, num=len(self._yvalues)) \
            * mw_abs_x_slope
        b_field_x = b_field_x_offset + \
            np.linspace(0, len(self._xvalues) - 1, num=len(self._xvalues)) * \
            b_field_x_slope
        self._xvalues = np.interp(mw_x, b_field_x, self._xvalues)

    def _extract_metadata_from_xml(self):
        # NOTE: Order of these statements is crucial not to overwrite values!
        xml_metadata = self.root[0][0][0].attrib
        xml_metadata.update(self.root[0][0].attrib)
        for childnode in self.root[0][0][0][0]:
            if 'Unit' in childnode.attrib:
                xml_metadata[childnode.attrib['Name']] = \
                    {'value': childnode.text, 'unit': childnode.attrib['Unit']}
            else:
                xml_metadata[childnode.attrib['Name']] = childnode.text
        self.xml_metadata = xml_metadata

    def _cut_data(self):
        self._get_magnetic_field_range()
        mask = (self._xvalues > self._bfrom) & (self._xvalues < self._bto)
        self._xvalues = self._xvalues[mask]
        self._yvalues = self._yvalues[mask]

    def _get_magnetic_field_range(self):
        """Get magnetic field range from preprocessed XML data."""
        self._bfrom = float(self.xml_metadata['Bfrom']['value'])
        self._bto = float(self.xml_metadata['Bto']['value'])

    def _hand_data_to_dataset(self):
        self.dataset.data.data = self._yvalues
        self.dataset.data.axes[0].values = self._xvalues
        self.dataset.data.axes[0].unit = 'mT'
        self.dataset.data.axes[0].quantity = 'magnetic field'
        self.dataset.data.axes[1].unit = 'mV'
        self.dataset.data.axes[1].quantity = 'intensity'

    def _infofile_exists(self):
        if self._get_infofile_name() and os.path.exists(
                self._get_infofile_name()[0]):
            return True
        print('No infofile found for dataset %s, import continued without '
              'infofile.' % os.path.split(self.source)[1])
        return False

    def _load_infofile(self):
        """Import infofile and parse it."""
        infofile_name = self._get_infofile_name()
        self._infofile.filename = infofile_name[0]
        self._infofile.parse()

    def _get_infofile_name(self):
        return glob.glob(self.source + '.info')

    def _assign_comment_as_annotation(self):
        comment = aspecd.annotation.Comment()
        comment.comment = self._infofile.parameters['COMMENT']
        self.dataset.annotate(comment)

    def _map_metadata(self, infofile_version):
        """Bring the metadata into a unified format."""
        mapper = aspecd.metadata.MetadataMapper()
        mapper.version = infofile_version
        mapper.metadata = self._infofile.parameters
        root_path = \
            os.path.split(os.path.split(os.path.abspath(__file__))[0])[0]
        mapper.recipe_filename = os.path.join(
            root_path, 'metadata_mapper_cwepr.yaml')
        mapper.map()
        self.dataset.metadata.from_dict(mapper.metadata)

    def _map_infofile(self):
        """Bring the metadata to a given format."""
        infofile_version = self._infofile.infofile_info['version']
        self._map_metadata(infofile_version)
        self._assign_comment_as_annotation()

    def _map_metadata_from_xml(self):
        self.dataset.metadata.temperature_control.temperature.value = \
            float(self.xml_metadata['Temperature']) + 273.15
        self.dataset.metadata.temperature_control.temperature.unit = 'K'
        self.dataset.metadata.experiment.type = \
            self.xml_metadata['KineticMode']
        self.dataset.metadata.experiment.runs = \
            self.xml_metadata['MeasurementCount']
        self.dataset.metadata.experiment.variable_parameter = \
            self.xml_metadata['XDatasource']
        self.dataset.metadata.spectrometer = \
            {'model': self.xml_metadata['Device'],
             'software': self.xml_metadata['SWV']}
        self.dataset.metadata.magnetic_field.start.from_string(
            self._dict_to_string(self.xml_metadata['Bfrom']))
        self.dataset.metadata.magnetic_field.stop.from_string(
            self._dict_to_string(self.xml_metadata['Bto']))
        self.dataset.metadata.magnetic_field.sweep_width.value = \
            float(self.xml_metadata['Bto']['value']) - \
            float(self.xml_metadata['Bfrom']['value'])
        self.dataset.metadata.magnetic_field.sweep_width.unit = \
            self.xml_metadata['Bfrom']['unit']
        self.dataset.metadata.magnetic_field.field_probe_type = 'Hall'
        self.dataset.metadata.magnetic_field.field_probe_model = 'builtin'
        if self._xvalues[-1] - self._xvalues[0] > 0:
            self.dataset.metadata.magnetic_field.sequence = 'up'
        else:
            self.dataset.metadata.magnetic_field.sequence = 'down'
        self.dataset.metadata.magnetic_field.controller = 'builtin'
        self.dataset.metadata.magnetic_field.power_supply = 'builtin'
        self.dataset.metadata.bridge.model = 'builtin'
        self.dataset.metadata.bridge.controller = 'builtin'
        self.dataset.metadata.bridge.power.from_string(self._dict_to_string(
            self.xml_metadata['MicrowavePower']))
        self.dataset.metadata.bridge.detection = 'mixer'
        self.dataset.metadata.bridge.frequency_counter = 'builtin'
        self.dataset.metadata.bridge.mw_frequency.value = \
            float(self.xml_metadata['MwFreq'])
        self.dataset.metadata.bridge.mw_frequency.unit = 'GHz'
        self.dataset.metadata.bridge.q_value = \
            float(self.xml_metadata['QFactor'])
        self.dataset.metadata.signal_channel.model = 'builtin'
        self.dataset.metadata.signal_channel.modulation_amplifier = 'builtin'
        self.dataset.metadata.signal_channel.accumulations = \
            int(self.xml_metadata['Accumulations'])
        self.dataset.metadata.signal_channel.modulation_frequency.from_string(
            self._dict_to_string(self.xml_metadata['ModulationFreq']))
        self.dataset.metadata.signal_channel.modulation_amplitude.from_string(
            self._dict_to_string(self.xml_metadata['Modulation']))
        self.dataset.metadata.signal_channel.phase = \
            float(self.xml_metadata['Phase'])
        self.dataset.metadata.probehead.model = 'builtin'
        self.dataset.metadata.probehead.coupling = 'critical'

    @staticmethod
    def _dict_to_string(dict_):
        return dict_['value'] + ' ' + dict_['unit']


class GoniometerSweepImporter(aspecd.io.DatasetImporter):
    """Import angular dependent data from Magnettech benchtop spectrometer.

    ..note::
        Metadata are only taken from the infofile, ignoring the (much likely
        more accurate) xml-file metadata.
    """

    def __init__(self, source=''):
        super().__init__(source=source)
        self._infofile = aspecd.infofile.Infofile()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.filenames = None
        self._data = None
        self._angles = []

    def _import(self):
        self._get_filenames()
        self._sort_filenames()
        self._import_all_spectra_to_list()
        self._hand_data_to_dataset()

        self._fill_axes()
        self._get_metadata()

    def _get_filenames(self):
        if not os.path.exists(self.source):
            raise FileNotFoundError
        self.filenames = glob.glob(os.path.join(self.source, '*[0-9]dg*.xml'))

    def _sort_filenames(self):

        def sort_key(string=''):
            num = string.split('gon_')[1]
            num = num.split('dg')[0]
            return int(num)

        self.filenames = sorted(self.filenames, key=sort_key)

    def _import_all_spectra_to_list(self):
        self._data = []
        # import all files without infofile
        for num, filename in enumerate(self.filenames):
            filename = filename[:-4]  # remove extension
            importer = cwepr.io.MagnettechXmlImporter(source=filename)
            importer.load_infofile = False
            self._data.append(cwepr.dataset.ExperimentalDataset())
            self._data[num].import_from(importer)
            self._angles.append(float(importer.xml_metadata['GonAngle']))
            for idx, angle in enumerate(self._angles):
                if angle > 359:
                    self._angles[idx] = 0
            # bring all measurements to the frequency of the first
            if num > 0:
                freq_correction = cwepr.processing.FrequencyCorrection()
                freq_correction.parameters['frequency'] = \
                    self._data[0].metadata.bridge.mw_frequency.value
                self._data[num].process(freq_correction)

            interpolate = cwepr.processing.AxisInterpolation()
            self._interpolation_to_same_number_of_points(interpolate, num)

    def _interpolation_to_same_number_of_points(self, interpolate, num):
        interpolate.parameters['points'] = len(self._data[0].data.data)
        self._data[num].process(interpolate)

    def _hand_data_to_dataset(self):
        my_array = np.ndarray((len(self._data[0].data.data), len(self._data)))
        self.dataset.data.data = my_array
        for num, set_ in enumerate(self._data):
            self.dataset.data.data[:, num] = set_.data.data

    def _fill_axes(self):
        self._fill_field_axis()
        self._fill_angle_axis()

    def _fill_field_axis(self):
        self.dataset.data.axes[0] = self._data[0].data.axes[0]

    def _fill_angle_axis(self):
        self.dataset.data.axes[1].values = np.asarray(self._angles)
        self.dataset.data.axes[1].unit = 'degree'
        self.dataset.data.axes[1].quantity = 'goniometer angle'

    def _get_metadata(self):
        """Import metadata from infofile.

        .. note::
            Currently, the metadata are only taken from the infofile.
            All information from the xml-file are completely ignored.

        """
        self._load_infofile()
        self._map_infofile()

    def _load_infofile(self):
        """Import infofile and parse it."""
        infofile_name = self._get_infofile_name()
        if not infofile_name:
            raise FileNotFoundError('Infofile not found')
        self._infofile.filename = infofile_name[0]
        self._infofile.parse()

    def _get_infofile_name(self):
        folder_path = os.path.split(self.source)[0]
        return glob.glob(folder_path + '.info')

    def _assign_comment_as_annotation(self):
        comment = aspecd.annotation.Comment()
        comment.comment = self._infofile.parameters['COMMENT']
        self.dataset.annotate(comment)

    def _map_metadata(self, infofile_version):
        """Bring the metadata into a unified format."""
        mapper = aspecd.metadata.MetadataMapper()
        mapper.version = infofile_version
        mapper.metadata = self._infofile.parameters
        root_path = \
            os.path.split(os.path.split(os.path.abspath(__file__))[0])[0]
        mapper.recipe_filename = os.path.join(
            root_path, 'metadata_mapper_cwepr.yaml')
        mapper.map()
        self.dataset.metadata.from_dict(mapper.metadata)

    def _map_infofile(self):
        """Bring the metadata to a given format."""
        infofile_version = self._infofile.infofile_info['version']
        self._map_metadata(infofile_version)
        self._assign_comment_as_annotation()



"""
Factory classes.

Currently, there is only one factory class:

  * :class:`cwepr.io.factory.DatasetImporterFactory`

    Class necessary in context of recipe-driven data analysis

"""


import os.path

import aspecd.io

from cwepr.io.bes3t import BES3TImporter
from cwepr.io.magnettech import MagnettechXmlImporter, GoniometerSweepImporter
from cwepr.io.txt_file import TxtImporter, CsvImporter
from cwepr.exceptions import NoMatchingFilePairError


class DatasetImporterFactory(aspecd.io.DatasetImporterFactory):
    """Factory for creating importer objects based on the source provided.

    Currently, the data formats are distinguished by their file extensions.

    """

    def __init__(self):
        super().__init__()
        self.supported_formats = {"BES3T": [".DTA", ".DSC"],
                                  "Magnettech": [".xml"],
                                  "Txt": [".txt"],
                                  "Csv": [".csv"]}
        self.importers_for_formats = {"BES3T": BES3TImporter,
                                      "Magnettech": MagnettechXmlImporter,
                                      "Txt": TxtImporter,
                                      'Csv': CsvImporter}
        self.data_format = None

    def _get_importer(self):
        """Main method returning the importer instance.

        Call the correct importer for the data format set. If no format is set,
        it is automatically determined from the given filename.

        Raises
        ------
        UnsupportedDataFormatError
            Raised if a format is set but does not match any of the supported
            formats

        """
        for end_ in [extension for sublist in self.supported_formats.values()
                     for extension in sublist]:
            if self.source.endswith(end_):
                self.source = self.source[:-len(end_)]
        if os.path.isdir(self.source):
            return GoniometerSweepImporter(source=self.source)
        self.data_format = self._find_format()
        special_importer = self.importers_for_formats[
            self.data_format](source=self.source)
        return special_importer

    def _find_format(self):
        """Find out the format of the given file.

        Determine the format of the given filename by checking if a data and
        metadata file matching any supported format are present.

        Determination is performed by checking if files with the correct name
        and extension are present.

        Raises
        ------
        NoMatchingFilePairError
            Raised if no source format doesn't match a supported format

        """
        for file_format, extensions in self.supported_formats.items():
            file_exists = []
            for extension in extensions:
                file_exists.append(os.path.isfile(self.source + extension))
            if all(file_exists):
                return file_format
        msg = "No file format was found for path: " + self.source
        raise NoMatchingFilePairError(message=msg)



"""
Importer for the Bruker BES3T format.

The Bruker BES3T format gets used by newer Bruker EPR spectrometers, namely
those running with either Xepr or Xenon.

The basic file format is comparably well documented, and it exists in at
least two different versions in the wild: 1.3 and 2.0.

But beware: The device-specific layer is *not* documented, at least not
according to a Bruker EPR Application Scientist. Hence, informed guessing is
necessary for those parts of the file, and it is not clear whether they are
similarly versioned as the basic format.
"""

import glob
import os
import re

import numpy as np

import aspecd.annotation
import aspecd.infofile
import aspecd.metadata
import aspecd.io
import aspecd.utils

import cwepr.metadata
import cwepr.exceptions


class BES3TImporter(aspecd.io.DatasetImporter):
    """Importer for the Bruker BES3T format.

    The Bruker BES3T format consists of at least two files, a data file with
    extension "DTA" and a descriptor file with extension "DSC". In case of
    multidimensional data, additional data files will be written (e.g.
    with extension ".YGF"), similarly to the case where the X axis is not
    equidistant (at least the BES3T specification allows this situation).

    This importer aims to take the parameters from the standard parameter
    layer if available, because it is given in SI units and is documented.

    """

    def __init__(self, source=None):
        # Dirty fix: Cut file extension
        if source.endswith((".DSC", ".DTA", ".YGF")):
            source = source[:-4]
        super().__init__(source=source)
        self.load_infofile = True
        # private properties
        self._infofile = aspecd.infofile.Infofile()
        self._dsc_dict = dict()
        self._mapper_filename = 'dsc_keys.yaml'
        self._is_two_dimensional = False
        self._dimensions = []
        self._file_encoding = ''
        self._points = int()

    def _import(self):
        self._extract_metadata_from_dsc()  # To get dimension information
        self._check_experiment()
        self._set_dataset_dimension()
        self._get_file_encoding()

        if self.load_infofile and self._infofile_exists():
            self._load_infofile()
            self._map_infofile()

        self._import_data()
        self._map_dsc_into_dataset()  # Needed for axes
        self._fill_axes()

        self._ensure_common_units()

    def _import_data(self):
        complete_filename = self.source + ".DTA"
        raw_data = np.fromfile(complete_filename, dtype=self._file_encoding)
        raw_data = np.reshape(raw_data, self._dimensions)
        if self._is_two_dimensional:
            raw_data = raw_data.T
        self.dataset.data.data = raw_data

    def _set_dataset_dimension(self):
        for key in ('YPTS', 'XPTS'):
            if key in self._dsc_dict.keys():
                self._dimensions.append(int(self._dsc_dict[key]))
        if len(self._dimensions) == 2:
            self._is_two_dimensional = True

    def _get_file_encoding(self):
        encodings = {
            'BIG': '>f8',
            'LIT': '<f8'
        }
        self._file_encoding = encodings[self._dsc_dict['BSEQ']]

    def _infofile_exists(self):
        if self._get_infofile_name() and os.path.exists(
                self._get_infofile_name()[0]):
            return True
        print('No infofile found for dataset %s, import continued without '
              'infofile.' % os.path.split(self.source)[1])
        return False

    def _extract_metadata_from_dsc(self):
        dsc_filename = self.source + '.DSC'
        with open(dsc_filename, 'r') as file:
            lines = file.read().splitlines()

        for line in lines:
            if not (line.startswith(('*', '#', '.')) or line == ''):
                if '\'' in line:
                    line = line.replace('\'', '')
                line = line.split(maxsplit=1)
                key = line[0]
                if len(line) > 1:
                    value = line[1]
                else:
                    value = ''
                if re.match(r'^[+-]?[0-9.]+([eE][+-]?[0-9]*)?$', value):
                    value = float(value)
                self._dsc_dict[key] = value

    def _map_dsc_into_dataset(self):
        yaml_file = aspecd.utils.Yaml()
        rootpath = os.path.split(os.path.abspath(__file__))[0]
        yaml_file.read_from(os.path.join(rootpath, self._mapper_filename))
        metadata_dict = {}
        metadata_dict = self._traverse(yaml_file.dict, metadata_dict)
        self.dataset.metadata.from_dict(metadata_dict)

    def _traverse(self, dict_, metadata_dict):
        for key, value in dict_.items():
            if isinstance(value, dict):
                metadata_dict[key] = {}
                self._traverse(value, metadata_dict[key])
            elif value in self._dsc_dict.keys():
                metadata_dict[key] = self._dsc_dict[value]
        return metadata_dict

    def _fill_axes(self):
        self._get_magnetic_field_axis()
        self.dataset.data.axes[0].quantity = 'magnetic field'
        self.dataset.data.axes[0].unit = self._dsc_dict['XUNI']
        self.dataset.data.axes[-1].quantity = 'intensity'

        if self._is_two_dimensional:
            self.dataset.data.axes[1].values = \
                np.fromfile(self.source + '.YGF', dtype=self._file_encoding)
            self.dataset.data.axes[1].quantity = self._dsc_dict['YNAM']
            self.dataset.data.axes[1].unit = self._dsc_dict['YUNI']

    def _get_magnetic_field_axis(self):
        # Abbreviations:
        start = self.dataset.metadata.magnetic_field.start.value
        points = int(self.dataset.metadata.magnetic_field.points)
        sweep_width = self.dataset.metadata.magnetic_field.sweep_width.value
        # because Bruker confounds number of steps and points
        stop = start + sweep_width - (sweep_width / (points + 1))
        # Set axis
        magnetic_field_axis = np.linspace(start, stop, points)
        assert len(magnetic_field_axis) == points, \
            'Length of magnetic field and number of points differ'
        assert len(magnetic_field_axis) == self.dataset.data.data.shape[0], \
            'Length of magnetic field and size of data differ'
        # set more values in dataset
        self.dataset.metadata.magnetic_field.stop.value = stop
        self.dataset.data.axes[0].values = magnetic_field_axis

    def _load_infofile(self):
        """Import infofile and parse it."""
        infofile_name = self._get_infofile_name()
        self._infofile.filename = infofile_name[0]
        self._infofile.parse()

    def _get_infofile_name(self):
        return glob.glob(''.join([self.source.strip(), '.info']))

    def _assign_comment_as_annotation(self):
        comment = aspecd.annotation.Comment()
        comment.comment = self._infofile.parameters['COMMENT']
        self.dataset.annotate(comment)

    def _map_metadata(self, infofile_version):
        """Bring the metadata into a unified format."""
        mapper = aspecd.metadata.MetadataMapper()
        mapper.version = infofile_version
        mapper.metadata = self._infofile.parameters
        root_path = \
            os.path.split(os.path.split(os.path.abspath(__file__))[0])[0]
        mapper.recipe_filename = os.path.join(
            root_path, 'metadata_mapper_cwepr.yaml')
        mapper.map()
        self.dataset.metadata.from_dict(mapper.metadata)

    def _map_infofile(self):
        """Bring the metadata to a given format."""
        infofile_version = self._infofile.infofile_info['version']
        self._map_metadata(infofile_version)
        self._assign_comment_as_annotation()

    def _ensure_common_units(self):
        """Transform selected values and units to common units.

        Because of information are doubled from the infofile and the
        DSC-file, some units are wrong and are corrected manually here.
        """
        # microwave frequency
        if self.dataset.metadata.bridge.mw_frequency.value > 50:
            self.dataset.metadata.bridge.mw_frequency.value /= 1e9
            self.dataset.metadata.bridge.mw_frequency.unit = 'GHz'
        # microwave power
        if self.dataset.metadata.bridge.power.value < 0.001:
            self.dataset.metadata.bridge.power.value *= 1e3
            self.dataset.metadata.bridge.power.unit = 'mW'
        # magnetic field objects
        objects_ = ('start', 'stop', 'sweep_width')
        for object_ in objects_:
            magnetic_field_object = getattr(
                self.dataset.metadata.magnetic_field, object_)
            if magnetic_field_object.unit == 'G':
                magnetic_field_object.value /= 10
                magnetic_field_object.unit = 'mT'
            setattr(
                self.dataset.metadata.magnetic_field, object_,
                magnetic_field_object)
        self.dataset.data.axes[0].values /= 10
        self.dataset.data.axes[0].unit = 'mT'
        # modulation frequency
        if self.dataset.metadata.signal_channel.modulation_frequency.value > \
                100:
            self.dataset.metadata.signal_channel.modulation_frequency.value \
                /= 1e3
            self.dataset.metadata.signal_channel.modulation_frequency.unit = \
                'kHz'
        self.dataset.metadata.signal_channel.modulation_amplitude.value *= 1e3
        self.dataset.metadata.signal_channel.modulation_amplitude.unit = 'mT'

    def _check_experiment(self):
        if self._dsc_dict['EXPT'] != 'CW':
            raise cwepr.exceptions.ExperimentTypeError(
                message='Experiment seems not to be a cw-Experiment.'
            )



"""Module containing the processing steps of the cwEPR package.

.. sidebar::
    processing *vs.* analysis

    For more details on the difference between processing and analysis,
    see the `ASpecD documentation <https://docs.aspecd.de/>`_.


A processing step always operates on a dataset and usually modifies the
numerical data contained therein. The result of a processing step is in any
case again a dataset, in contrast to analysis steps where this is not
necessarily the case. Typical routine processing steps are normalisation (to
area, amplitude, maximum, minimum), and for EPR spectroscopy such things as
field and frequency correction.


Processing steps specific for cw-EPR data
-----------------------------------------

Currently, the following processing steps are implemented:

  * :class:`FieldCorrection`
  * :class:`FrequencyCorrection`
  * :class:`GAxisCreation`

  * :class:`BaselineCorrectionWithPolynomial`

  * :class:`NormalisationOfDerivativeToArea`
  * :class:`Normalisation`

  * :class:`Integration`

  * :class:`Averaging2DDataset`
  * :class:`SubtractVector`

Implemented but not working as they should:

  * :class:`PhaseCorrection`
  * :class:`AutomaticPhaseCorrection`


General processing steps inherited from the ASpecD framework
------------------------------------------------------------

Besides the processing steps specific for TREPR data, a number of further
processing steps that are generally applicable to spectroscopic data have
been inherited from the underlying ASpecD framework:

* :class:`ScalarAlgebra`

  Perform scalar algebraic operation on one dataset.

  Operations available: add, subtract, multiply, divide (by given scalar)

* :class:`ScalarAxisAlgebra`

  Perform scalar algebraic operation on axis values of a dataset.

  Operations available: add, subtract, multiply, divide, power (by given scalar)

* :class:`DatasetAlgebra`

  Perform scalar algebraic operation on two datasets.

  Operations available: add, subtract

* :class:`Projection`

  Project data, *i.e.* reduce dimensions along one axis.

* :class:`SliceExtraction`

  Extract slice along one ore more dimensions from dataset.

* :class:`BaselineCorrection`

  Correct baseline of dataset.

* :class:`Averaging`

  Average data over given range along given axis.

* :class:`Filtering`

  Filter data


Further processing steps implemented in the ASpecD framework can be used as
well, by importing the respective modules. In case of recipe-driven data
analysis, simply prefix the kind with ``aspecd``:

.. code-block:: yaml

    - kind: aspecd.processing
      type: <ClassNameOfProcessingStep>


Categories of processing steps
==============================

Processing steps can be categorised further. The following is an attempt to
do that for cwEPR data and at the same time a list of processing steps one
would like to have implemented. Besides that, it seems that this list
evolves more and more towards a summary of how to properly record and
(post-)process cwEPR data.

For more authoritative answers, you may as well have a look into the EPR
literature, particularly the "EPR Primer" by Chechik/Carter/Murphy and
the book on quantitative EPR by the Eatons.

.. todo::
    Add references here, using the BIBTeX plugin


Corrections
-----------

When analysing cwEPR data, usually, a series of simple correction steps is
performed prior to any further analysis. This is particularly important if
you plan to compare different datasets or if you would like to compare your
spectra with those from the literature (always a good idea, though).

  * Magnetic field correction

    Usually, the magnetic field in an EPR measurement needs to be determined
    by measuring a field standard in the identical setup, as the actual
    magnetic field at the sample will usually differ from the field set in
    the software.

    Appropriate magnetic field correction becomes particularly important if
    you are interested in absolute *g* values of your sample, *e.g.* to
    compare it to literature data or quantum-chemical calculations and to
    get ideas as to where the unpaired spin may predominantly reside on (in
    terms of nuclear species).

  * Microwave frequency correction

    Comparing datasets is only possible in a meaningful manner if they are
    either corrected for same frequency, or their magnetic field axes
    converted in a g axis.

  * Microwave phase correction

    Usually, cwEPR spectra are not recorded with quadrature detection,
    *i.e.*, with both, absorptive and dispersive signal components. However,
    using the Hilbert transform, one can reconstruct the dispersive signal (
    imaginary component) and correct the phase of the microwave source this way.

  * Baseline correction

    However careful measurements are performed, baselines are quite often
    encountered. There are two different kinds of baseline that need to be
    corrected in different ways. Drifts of some kind can usually be handled
    by fitting and afterwards subtracting a (low-order) polynomial to the data.

    Particularly for low-temperature data, weak signals, and large magnetic
    field sweep ranges, resonator background can become quite dramatic.
    Here, usually the only viable way is to record the empty resonator
    independently under as much identical conditions as possible compared to
    recording the signal of the actual sample (but with slightly broader
    field range to compensate for different microwave frequency) and
    afterwards subtracting this dataset (empty resonator, *i.e.* resonator
    background signal) from the signal of the actual sample.


Algebra
-------

.. sidebar::
    Availability

    Algebra is available with:
    :class:`ScalarAlgebra` and
    :class:`DatasetAlgebra`

Comparing datasets often involves adding, subtracting, multiplying or
dividing the intensity values by a given fixed number. Possible scenarios
where one wants to multiply the intensity values of a cwEPR spectrum may be
comparing spectra resulting from a single species from those of known two
species, different (known) concentrations and alike.

Of course, dividing the intensity of the spectrum by the maximum intensity
is another option, however, this would be normalisation to maximum (not
always a good idea, usually normalising to area or amplitude is better),
and this is handled by a different set of processing steps (see below).

This type of simple algebra is quite *different* from adding or subtracting
datasets together. Whereas simple algebra really is a one-liner in terms of
implementation, handling different datasets involves ensuring commensurable
axis dimensions and ranges, to say the least. Dataset algebra is available as
well in this module.


Normalisation
-------------

Normalising data to some common characteristic is a prerequisite for
comparing datasets among each other.

There is a number of normalisations that are common for nearly every kind of
data, and as such, these normalisation steps should probably eventually be
implemented within the ASpecD framework. As there are:

  * Normalisation to maximum

    Simply divide the intensity values by their maximum

    Often used as a very simple "normalisation" approach. Depends highly on
    the situation and focus of the representation, but usually,
    other methods such as normalisation to amplitude or area, are better
    suited.

  * Normalisation to minimum

    Simply divide the intensity values by their minimum

    The same as for the normalisation to maximum applies here. Furthermore,
    normalising to the minimum usually only makes sense in case of
    prominent negative signal components, as in first-derivative spectra in
    cwEPR spectroscopy.

  * Normalisation to amplitude

    Divide the intensity values by the absolute of the difference between
    maximum and minimum intensity value

    Usually better suited as a simple normalisation than the naive
    normalising to maximum or minimum described above. However, it strongly
    depends on what you are interested in comparing and want to highlight.

  * Normalisation to area

    Divide the intensity values by the area under the curve of the spectrum

    Not as easy as it looks like for first-derivative cwEPR spectra,
    as here, you are usually interested in normalising to the same area (
    *i.e.*, integral of the curve) of the absorptive (zeroth-derivative or
    zeroth harmonic) spectrum.

    At least given appropriate measurement conditions (no saturation, no line
    broadening due to overmodulation, proper phasing), the cwEPR signal
    intensity should be proportional to the number of spins in the active
    volume of the resonator/probehead. Therefore, with all crucial
    experimental parameters directly affecting the signal strength being
    equal (microwave power, modulation amplitude), normalising to same area
    should be the most straight-forward way of comparing two spectra in a
    meaningful way.

    Bare in mind, however, that spectra with strongly different overall line
    width will have dramatically different minima and maxima, making
    comparison of this kind sometimes less meaningful.


Besides these rather general ways of normalising spectra (although described
above particularly with cwEPR data in mind), there are some other
normalisations more particular to cwEPR spectroscopy:

  * Normalisation to same number of scans

    Some spectrometers (probably only older ones) did usually sum the
    intensity for each scan, rather than afterwards dividing by the number
    of scans, making comparison of spectra with different number of scans
    quite tricky.

    Make sure you know exactly what you do before applying (or not applying)
    such normalisation if you would like to do some kind of (semi-)quantitative
    analysis of your data.

  * Normalisation to same receiver gain

    The preamplifiers in the signal channel (as the digitising unit in cwEPR
    spectrometers is usually called) have usually a gain that can be
    adjusted to the signal strength of the actual sample. Of course,
    this setting will have a direct impact on the intensity values recorded (
    usually something like mV).

    Comparing spectra recorded with different receiver gain settings
    therefore requires the user to *first* normalise the data to the same
    receiver gain setting. Otherwise, (semi-)quantiative comparison is not
    possible and will lead to wrong conclusions.

    Note on the side: Adjusting the receiver gain for each measurement is
    highly recommended, as setting it too high will make the signal clip and
    distort the signal shape, and setting it too low will result in data
    with (unnecessary) poor signal-to-noise ratio.



Working with 2D datasets
------------------------

2D datasets in cwEPR spectroscopy, huh? Well, yes, more often than one might
expect in the beginning. There are the usual suspects such as power sweeps
and modulation amplitude sweeps, each varying (automatically) one parameter
in a given range and record spectra for each value.

There are, however, other types of 2D datasets that are quite useful in
cwEPR spectroscopy. Some vendors of EPR spectrometers offer no simple way of
saving each individual scan in a series of accumulations. However, this may
sometimes be of interest, particularly as a single "spike" due to some
external event or other malfunctioning may otherwise ruin your entire
dataset, however long it might have taken to record it. Therefore, one way
around this limitations is to perform a 2D experiment with repeated field
scans, but saving each scan as a row in a 2D dataset.

Generally, there are at least two different processing steps of interest for
2D datasets:

  * Projection along one axis

    Equivalent to averaging along that axis

    If recording multiple scans of one and the same spectrum for better
    signal-to-noise ratio, but saving each scan individually within a row of
    a 2D dataset, this is the way to get the dataset with improved
    signal-to-noise ratio originally intended.

    May as well be used for rotation patterns, *i.e.*, angular-dependent
    measurements, if there turns out to be no angular dependence in the
    data. In this case, at least you save the measurement time by having a
    dataset with clearly better signal-to-noise ratio than initially intended.

  * Extraction of a slice along one dimension

    Having a 2D dataset, we may often be interested in only one slice along
    one dimension.

    Typical examples would be comparing two positions of the goniometer
    (zero and 180 degree would be an obvious choice) or slices with similar
    parameters for different datasets.


More complicated and probably more involved processing of 2D datasets would
be to (manually) inspect the individual scans and decide which of those to
average, *e.g.* in case of one problematic scan in between, be it due to
external noise sources or spectrometer problems.


Handling multiple datasets
--------------------------

Comparing multiple datasets by plotting them in one and the same axis is a
rather simple way of handling multiple datasets. However, usually, you would
like to perform much more advanced operations on multiple datasets, such as
adding and subtracting one from the other.

May sound pretty simple at first, but is indeed pretty demanding in terms of
its implementation, as internally, you need to check for quite a number of
things, such as commensurable axes and ranges. However, this is a rather
general problem of all kinds of datasets, hence it may be that this
functionality eventually gets incorporated in the ASpecD framework.

Particularly in EPR spectroscopy, each measurement will have a unique
microwave frequency for which the data were recorded. Therefore, to combine
the numerical values of two datasets (subtract, add, average), you will
first need to correct them for same microwave frequency. This will generally
result in different field axes for different datasets. Furthermore,
some vendors like to record data with non-equidistant field axes as well,
making handling of those datasets additionally messy.

  * Subtract a dataset from another dataset

    Ensure the datasets are compatible in terms of their axes (dimension,
    quantity, unit, common area of values), subtract the common range of
    values and return only the subtracted (*i.e.*, usually truncated) dataset.

    A common use case for subtracting a dataset from another would be a
    resonator background signal independently recorded, or some other
    background signal such as the "glass signal" (from impurities in the
    glass tube you've used).

    Other, more advanced applications may involve subtracting the spectrum
    of a single species from that of a spectrum consisting this and other
    species. However, in such case be aware of the fact that the spectrum
    containing more than one species may not be a simple superposition of
    the spectra of the two independent species.

  * Add a dataset to another dataset

    Ensure the datasets are compatible in terms of their axes (dimension,
    quantity, unit, common area of values), add the common range of values
    together and return only the summed (*i.e.*, usually truncated) dataset.

  * Average two datasets

    Ensure the datasets are compatible in terms of their axes (dimension,
    quantity, unit, common area of values), average the common range of values
    together and return only the averaged (*i.e.*, usually truncated) dataset.

    A common use case if you performed several independent measurements of
    the same sample (with otherwise similar/comparable parameters) and would
    like to average for better signal-to-noise.


Other processing steps
----------------------

There may well be further types of processing steps the authors are currently
not aware of or didn't dare to document here.


Note to developers
==================

Processing steps can be based on analysis steps, but not inverse! Otherwise,
we get cyclic dependencies what should obviously be avoided in order to keep
code working.

Implementing own processing steps is rather straight-forward. For details,
see the documentation of the :mod:`aspecd.processing` module.


Module documentation
====================

What  follows is the API documentation of each class implemented in this module.

"""
import copy
import math
import numpy as np
import scipy.constants
import scipy.integrate
import scipy.interpolate
import scipy.signal

import aspecd.processing
import cwepr.exceptions


class FieldCorrection(aspecd.processing.SingleProcessingStep):
    """Processing step for field correction.

    Perform a linear field correction of the data with a correction value
    previously determined.

    .. todo::
        Doublecheck for units.

    Attributes
    ----------
    parameters['correction_value']: :class:`float`
        Previously determined correction value.

    """

    def __init__(self):
        super().__init__()
        self.parameters["correction_value"] = None
        self.description = "Linear field correction"

    def _perform_task(self):
        """Shift all field axis data points by the correction value."""
        for axis in self.dataset.data.axes:
            # TODO: Question: Better check for quantity rather than unit? (
            #   Difficult if not filled)
            # if axis.quantity == 'magnetic field'
            if axis.unit in ('mT', 'G'):
                self.dataset.data.axes[0].values += \
                    self.parameters["correction_value"]


class FrequencyCorrection(aspecd.processing.SingleProcessingStep):
    """Convert data of a given frequency to another given frequency.

    This is used to make spectra comparable.

    Attributes
    ----------
    self.parameters['frequency']
        Frequency to correct for.

        Default: 9.5

    """

    def __init__(self):
        super().__init__()
        self.parameters["frequency"] = 9.5
        self.description = "Correct magnetic field axis for given frequency"

    def _perform_task(self):
        """Perform the actual transformation / correction.

        For the conversion the x axis data is first converted to an axis in
        units of using the given frequency, then converted back using target
        frequency.
        """
        nu_target = self.parameters['frequency']
        for axis in self.dataset.data.axes:
            # TODO: Question: Better check for quantity rather than unit? (
            #   Difficult if not filled)
            # if axis.quantity == 'magnetic field'
            if axis.unit in ('mT', 'G'):
                axis.values = self._correct_field_for_frequency(nu_target,
                                                                axis.values)
                self._write_new_frequency()

    def _correct_field_for_frequency(self, nu_target=None,
                                     b_initial=None):
        """
        Calculate new field axis for given frequency.

        Parameters
        ----------
        nu_target : :class:`float`
            Frequency the magnetic field should be computed for

        b_initial : :class:`numpy.ndarray`
            Original field axis

        Returns
        -------
        b_target : :class:`numpy.ndarray`
            Computed field axis

        """
        nu_initial = self.dataset.metadata.bridge.mw_frequency.value
        b_target = (nu_target / nu_initial) * b_initial
        return b_target

    def _write_new_frequency(self):
        self.dataset.metadata.bridge.mw_frequency.value = \
            self.parameters['frequency']


class GAxisCreation(aspecd.processing.SingleProcessingStep):
    """Change magnetic field axis to g axis.

    Attributes
    ----------
    self.parameters['frequency']
        frequency to calculate g axis with.

    """

    def __init__(self):
        super().__init__()
        self.parameters["frequency"] = 9.5
        self.description = "Return a g-axis."

    def _perform_task(self):
        for axis in self.dataset.data.axes:
            if axis.unit in ('mT', 'G'):
                axis.values = self._create_g_axis(axis.values)
                axis.unit = ''

    def _create_g_axis(self, field_values=None):
        planck_constant = scipy.constants.value('Planck constant')
        mu_b = scipy.constants.value('electron mag. mom.')
        # pylint: disable=invalid-name
        nu = self.dataset.metadata.bridge.mw_frequency.value
        g_ = (planck_constant * nu) / (mu_b * field_values)
        return g_


class BaselineCorrectionWithPolynomial(aspecd.processing.SingleProcessingStep):
    """Perform a baseline correction assuming an underlying polynomial function.

    The coefficients to use will be calculated using the given order  and
    written in the parameters. If no order is explicitely given, a shifted
    baseline of zeroth order is assumed and will be processed for.
    See also: :class:`cwepr.analysis.BaselineCorrectionWithCalculatedDataset`.

    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        percentage :
            Parts of the spectrum to be considered as baseline, can be given
            as list or single number. If one number is given, it takes that
            percentage from both sides, respectively, i.e. 10 means 10% left
            and 10 % right. If a list of two numbers is provided,
            the corresponding percentages are taken from each side of the
            spectrum, i.e. ``[5, 20]`` takes 5% from the left side and 20%
            from the right.

            Default: [10, 10]

        order : :class:`int`
            The order for the baseline correction if no coefficients are given.

            Default: 0

        coefficients:
            Filled during evaluation of the task, coefficients of the
            baseline polynomial.

    """

    def __init__(self):
        super().__init__()
        self.description = "Subtraction of baseline polynomial"
        self.parameters['percentage'] = [10, 10]
        self.parameters['order'] = 0
        self._cut_x_data = np.ndarray([])
        self._cut_y_data = np.ndarray([])
        self.parameters['coefficients'] = None

    @staticmethod
    def applicable(dataset):  # noqa: D102
        return dataset.data.data.ndim == 1

    def _sanitise_parameters(self):
        if isinstance(self.parameters['percentage'], (float, int)):
            percentage = self.parameters['percentage']
            self.parameters['percentage'] = [percentage, percentage]
        if isinstance(self.parameters['percentage'], list) and len(
                self.parameters['percentage']) == 1:
            percentage = self.parameters['percentage'][0]
            self.parameters['percentage'] = [percentage, percentage]

    def _perform_task(self):
        """Perform the actual correction.

        Baseline correction is performed by subtraction of  an evaluated
        polynomial.
        """
        self._get_spectrum_to_evaluate()
        values_to_subtract = self._get_values_to_subtract()
        self.dataset.data.data -= values_to_subtract

    def _get_spectrum_to_evaluate(self):
        number_of_points = len(self.dataset.data.data)
        points_left = math.ceil(number_of_points * self.parameters[
            "percentage"][0] / 100.0)
        points_right = math.ceil(number_of_points * self.parameters[
            "percentage"][1] / 100.0)
        data = self.dataset.data.data
        x_axis = self.dataset.data.axes[0].values
        self._cut_y_data = np.r_[data[:points_left], data[-points_right:]]
        self._cut_x_data = np.r_[x_axis[:points_left], x_axis[-points_right:]]

    def _get_values_to_subtract(self):
        # pylint: disable=using-constant-test
        if np.polynomial.Polynomial:
            polynomial_ = np.polynomial.Polynomial.fit(self._cut_x_data,
                                                       self._cut_y_data,
                                                       self.parameters['order'])
            self.parameters['coefficients'] = polynomial_.coef
            return polynomial_(self.dataset.data.axes[0].values)
        polynomial_ = np.polyfit(self._cut_x_data, self._cut_y_data,
                                 self.parameters['order'])
        self.parameters['coefficients'] = polynomial_
        return np.polyval(polynomial_, self.dataset.data.axes[0].values)


class PhaseCorrection(aspecd.processing.SingleProcessingStep):
    """Phase correction if phase angle is given directly or in metadata.

    Therefore the here used implementation of the processing step is also
    highly problematic as most phase deviation is not introduced on purpose
    and listed in the metadata.

    Attributes
    ----------
    parameters : :class:`dict`
        phase_angle_value: :class:`float`
            Value of the phase shift, can be given in Degree or Rad

        phase_angle_unit: :class:`str`
            Unit of the phase shift.

            Default: deg

    """

    def __init__(self):
        super().__init__()
        self.description = "Phase Correction via Hilbert transform"
        self.parameters["phase_angle_value"] = None
        self.parameters["phase_angle_unit"] = 'deg'

    def _perform_task(self):
        """Perform the actual phase correction.

        The phase angle is acquired from the dataset's metadata and
        transformed to radians if necessary. The phase correction is then
        applied and the corrected data inserted into the dataset.
        """
        if not self.parameters["phase_angle_value"]:
            self._get_phase_from_metadata()
        self._convert_deg_to_rad()
        self._do_phase_correction()

    def _do_phase_correction(self):
        data = self.dataset.data.data
        analytic_signal = scipy.signal.hilbert(data)
        corrected_analytic_signal = np.exp(-1j * self.parameters[
            "phase_angle_value"]) * analytic_signal 
        corrected_data = np.real(corrected_analytic_signal)
        self.dataset.data.data = corrected_data

    def _convert_deg_to_rad(self):
        if self.parameters["phase_angle_unit"] == "deg":
            self.parameters["phase_angle_value"] = \
                (np.pi * self.parameters["phase_angle_value"]) / 180
            self.parameters["phase_angle_unit"] = "rad"

    def _get_phase_from_metadata(self):
        phase_angle_raw = self.dataset.metadata.signal_channel.phase
        self.parameters["phase_angle_value"] = phase_angle_raw.value
        self.parameters["phase_angle_unit"] = phase_angle_raw.unit


class AutomaticPhaseCorrection(aspecd.processing.SingleProcessingStep):
    """Automatic phase correction via Hilbert transform.

    ..important::
        Experimental state: Other methods have been proven to provide a
        better and reliable phase correction.

    .. todo::
        Does not work properly. Already gives wrong values with simulated
        data without a hyperfine-coupling. Reimplement with other method...

    Adapted from the matlab functionality in the cwEPR-toolbox.
    """

    def __init__(self):
        super().__init__()
        # Public properties
        self.description = "Automatic phase correction via Hilbert transform"
        self.parameters['order'] = 1
        self.parameters['points_percentage'] = 10
        self.parameters['phase_angle'] = 0
        # private properties
        self._analytic_signal = None
        self._points_per_side = None
        self._area_under_curve = None

    def _perform_task(self):
        self._analytic_signal = scipy.signal.hilbert(self.dataset.data.data)
        self._find_initial_negative_area()
        self._find_best_phase()
        self._reconstruct_real_signal()
        self._print_results_to_command_line()

    def _find_initial_negative_area(self):
        """Get area/values below zero as indicator of the phase deviation."""
        ft_sig_tmp = self._analytic_signal
        if self.parameters['order'] > 0:
            # pylint: disable=unused-variable
            for j in range(self.parameters['order']):  # integrate j times
                ft_sig_tmp = scipy.integrate.cumtrapz(self._analytic_signal,
                                                      initial=0)
        ft_sig_tmp = self._baseline_correction(signal=np.real(ft_sig_tmp))
        elements_inf_zero = [x for x in ft_sig_tmp if x < 0]
        self._area_under_curve = abs(np.trapz(elements_inf_zero))

    def _baseline_correction(self, signal=None):
        signal = np.asarray(signal)
        signal_size = signal.size
        if len(signal.shape) > 1 and signal.shape[1] != 1:
            signal.transpose()
        self._points_per_side = \
            int(np.ceil((self.parameters['points_percentage'] / 100) *
                        signal_size))

        data_parts = self._extract_points(signal)
        x_axis_parts = self._extract_points(self.dataset.data.axes[0].values)
        coefficients = np.polyfit(x_axis_parts, data_parts,
                                  deg=self.parameters['order'])
        baseline = np.polyval(coefficients, self.dataset.data.axes[0].values)

        corrected_signal = signal - baseline
        if len(signal.shape) > 1 and signal.shape[1] != 1:
            corrected_signal = corrected_signal.transpose()
        return corrected_signal

    def _extract_points(self, values):
        # pylint: disable=invalid-unary-operand-type
        vector_parts = np.concatenate([values[:self._points_per_side],
                                       values[-self._points_per_side:]])
        return vector_parts

    def _find_best_phase(self):
        min_angle = -np.pi / 2
        max_angle = np.pi / 2
        # TODO: introduce parameter step width/number of points.
        angles = np.linspace(min_angle, max_angle, num=181)
        for angle in angles:
            rotated_signal = (np.exp(1j * angle) * self._analytic_signal)
            if self.parameters['order'] > 0:
                # pylint: disable=unused-variable
                for j in range(self.parameters['order']):
                    rotated_signal = scipy.integrate.cumtrapz(rotated_signal,
                                                              initial=0)
            rotated_signal = self._baseline_correction(signal=np.real(
                rotated_signal))
            elements_inf_zero = [x for x in rotated_signal if x < 0]
            area = abs(np.trapz(elements_inf_zero))
            if area < self._area_under_curve:
                self._area_under_curve = area
                self.parameters['phase_angle'] = angle

    def _reconstruct_real_signal(self):
        self.dataset.data.data = np.real(np.exp(1j * self.parameters[
            'phase_angle']) * self._analytic_signal)
        assert not np.iscomplex(self.dataset.data.data).all()

    def _print_results_to_command_line(self):
        phi_degree = self.parameters['phase_angle'] * 180 / np.pi
        print('Phase correction was done with phi = %.3f degree' % phi_degree)


class NormalisationOfDerivativeToArea(aspecd.processing.SingleProcessingStep):
    """Normalise a spectrum to the area under the curve.

    As typical cw-EPR spectra are derivative spectra, calculating the area
    under the curve involves an integration step beforehand. This is done
    here as well.

    .. note::
        If the integrated spectra has a baseline shift, it is not currently
        accounted for.
    """

    def __init__(self):
        super().__init__()
        self.description = "Normalisation to area"
        self.undoable = True
        self._area = float()

    def _perform_task(self):
        self._integrate_spectrum()
        self.dataset.data.data /= self._area

    def _integrate_spectrum(self):
        integrated_spectrum = \
            scipy.integrate.cumtrapz(self.dataset.data.data, initial=0)
        self._area = np.trapz(integrated_spectrum)


class Normalisation(aspecd.processing.Normalisation):
    """Normalise data.

    For an extended documentation of the kinds implemented directly in
    ASpecD, see the corresponding documentation:
    :class:`aspecd.processing.Normalisation`.

    The basic usage is as follows:

    .. code-block:: yaml

       - kind: processing
         type: Normalisation
         properties:
           parameters:
             kind: receiver_gain


    The two additional kinds are explained below.

    Due to the logarithmic scale of the *receiver gain* (in dB) at least in
    BRUKER spectrometers, it has to be transferred into the "normal" scale. It
    calculates as following:

        receiver gain = 10^(receiver gain in dB/20)

    Source: Stefan Stoll, EasySpin source code, according to Xenon Manual 2011

    The normalisation according to the *number of scans* is necessary to make
    spectra in which the intensity of different scans is the sum of the
    single scans, comparable to those where it is averaged.

    .. important::
        Know what you are doing, sometimes the spectrometer's software does this
        step silently...
    """

    def _perform_task(self):
        super()._perform_task()
        if 'receiver' in self.parameters["kind"].lower():
            self._normalise_for_receiver_gain()
        elif 'scan_number' in self.parameters["kind"].lower():
            self.dataset.data.data /= \
                self.dataset.metadata.signal_channel.accumulations
        else:
            # todo: If only python >=3.6: Use first version
            #raise ValueError(f'Kind {self.parameters["kind"]} not recognised.')
            raise ValueError('Kind %s not recognised.' % self.parameters[
                "kind"])

    def _normalise_for_receiver_gain(self):
        receiver_gain = self.dataset.metadata.signal_channel.receiver_gain.value
        receiver_gain_value = 10 ** (receiver_gain / 20)
        self.dataset.data.data /= receiver_gain_value


class Integration(aspecd.processing.SingleProcessingStep):
    """Perform an indefinite integration.

    Indefinite integration means integration yielding an integral function.
    The quality of the integration can be determined using
    :class:`cwepr.analysis.IntegrationVerification`
    """

    def __init__(self):
        super().__init__()
        self.description = "Indefinite Integration"

    def _perform_task(self):
        """Perform the actual integration.

        Perform the actual integration using trapezoidal integration
        functionality from scipy. The keyword argument initial=0 is used to
        yield a list of length identical to the original one.
        """
        x_coordinates = self.dataset.data.axes[0].values
        y_coordinates = self.dataset.data.data

        integral_values = \
            scipy.integrate.cumtrapz(y_coordinates, x_coordinates, initial=0)
        self.dataset.data.data = integral_values


class AxisInterpolation(aspecd.processing.SingleProcessingStep):
    """Interpolating axes to given number of equidistant field points.

    Iterates over axes and takes the first axis that is not equidistant and
    interpolates it and the data as well.

    Attributes
    ----------
    parameters['points']
        Number of points that should be interpolated to.

    """

    def __init__(self):
        super().__init__()
        self.description = 'Interpolate magnetic field axis to get ' \
                           'equidistant field points.'
        self.parameters['points'] = None

    def _perform_task(self):
        for num, axis in enumerate(self.dataset.data.axes):
            if not axis.equidistant:
                if not self.parameters['points']:
                    self._get_axis_length(ax_nr=num)
                self._interpolate_axis(num)
                break

    def _interpolate_axis(self, ax_number=None):
        points = self.parameters['points']
        # Actual interpolation
        start = self.dataset.metadata.magnetic_field.start.value
        stop = self.dataset.metadata.magnetic_field.stop.value
        new_x_axis = np.linspace(start, stop, num=points)
        self.dataset.data.data = np.interp(new_x_axis, self.dataset.data.axes[
            ax_number].values, self.dataset.data.data)
        self.dataset.data.axes[ax_number].values = new_x_axis

    def _get_axis_length(self, ax_nr):
        self.parameters['points'] = len(self.dataset.data.axes[ax_nr].values)


class Averaging2DDataset(aspecd.processing.SingleProcessingStep):
    """Average over 2D dataset to get one dimensional dataset.

    Attributes
    ----------
    parameters['axis']
        Axis along which should be averaged.

        Default: 1

    """

    def __init__(self):
        super().__init__()
        self.parameters['axis'] = 1
        self.description = 'Project 2D data in one dimension.'
        self.undoable = True

    def _perform_task(self):
        old_dataset = copy.deepcopy(self.dataset)
        self.dataset.data.data = np.average(self.dataset.data.data,
                                            axis=self.parameters['axis'])
        if self.parameters['axis'] == 1:
            self.dataset.data.axes[1] = old_dataset.data.axes[2]
            del self.dataset.data.axes[2]

    @staticmethod
    def applicable(dataset):  # noqa: D102
        return len(dataset.data.axes) == 3


class SubtractVector(aspecd.processing.SingleProcessingStep):
    """Subtract vector of same length from dataset.

    With a 2D dataset, the vector is subtracted from

    Attributes
    ----------
    parameters['vector']
        Vector that is subtracted from the data

    """

    def __init__(self):
        super().__init__()
        self.description = 'Subtract vector of same length from dataset.'
        self.parameters['vector'] = []

    def _perform_task(self):
        if len(self.parameters['vector']) != \
                self.dataset.data.data.shape[0]:
            raise cwepr.exceptions.DimensionError(
                message='Vector to subtract is not of the same length as '
                        'dataset.')
        if self.dataset.data.data.ndim == 1:
            self.dataset.data.data -= self.parameters['vector']
        elif self.dataset.data.data.ndim == 2:
            for second_dim in range(self.dataset.data.data.shape[1]):
                self.dataset.data.data[:, second_dim] -= \
                    self.parameters['vector']
        else:
            raise cwepr.exceptions.DimensionError(
                message='Dataset has weird shape.')


class ScalarAlgebra(aspecd.processing.ScalarAlgebra):
    """Perform scalar algebraic operation on one dataset.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.ScalarAlgebra`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In case you would like to add a fixed value of 42 to your dataset:

    .. code-block:: yaml

       - kind: processing
         type: ScalarAlgebra
         properties:
           parameters:
             kind: add
             value: 42

    Similarly, you could use "minus", "times", "by", "add", "subtract",
    "multiply", or "divide" as kind - resulting in the given algebraic
    operation.

    """


class ScalarAxisAlgebra(aspecd.processing.ScalarAxisAlgebra):
    """Perform scalar algebraic operation on the axis of a dataset.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.ScalarAxisAlgebra`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In case you would like to add a fixed value of 42 to the first axis
    (index 0) your dataset:

    .. code-block:: yaml

       - kind: processing
         type: ScalarAxisAlgebra
         properties:
           parameters:
             kind: plus
             axis: 0
             value: 42

    Similarly, you could use "minus", "times", "by", "add", "subtract",
    "multiply", "divide", and "power" as kind - resulting in the given
    algebraic operation.

    """


class DatasetAlgebra(aspecd.processing.DatasetAlgebra):
    """Perform scalar algebraic operation on two datasets.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.DatasetAlgebra`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In case you would like to add the data of the dataset referred to by its
    label ``label_to_other_dataset`` to your dataset:

    .. code-block:: yaml

       - kind: processing
         type: DatasetAlgebra
         properties:
           parameters:
             kind: plus
             dataset: label_to_other_dataset

    Similarly, you could use "minus", "add", "subtract" as kind - resulting
    in the given algebraic operation.

    As mentioned already, the data of both datasets need to have identical
    shape, and comparison is only meaningful if the axes are compatible as
    well. Hence, you will usually want to perform a CommonRangeExtraction
    processing step before doing algebra with two datasets:

    .. code-block:: yaml

       - kind: multiprocessing
         type: CommonRangeExtraction
         results:
           - label_to_dataset
           - label_to_other_dataset

       - kind: processing
         type: DatasetAlgebra
         properties:
           parameters:
             kind: plus
             dataset: label_to_other_dataset
         apply_to:
           - label_to_dataset

    """


class Projection(aspecd.processing.Projection):
    """Project data, *i.e.* reduce dimensions along one axis.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.Projection`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In the simplest case, just invoke the projection with default values:

    .. code-block:: yaml

       - kind: processing
         type: Projection

    This will project the data along the first axis (index 0), yielding a 1D
    dataset.

    If you would like to project along the second axis (index 1), simply set
    the appropriate parameter:

    .. code-block:: yaml

       - kind: processing
         type: Projection
         properties:
           parameters:
             axis: 1

    This will project the data along the second axis (index 1), yielding a 1D
    dataset.

    """


class SliceExtraction(aspecd.processing.SliceExtraction):
    """Extract slice along one ore more dimensions from dataset.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.SliceExtraction`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In the simplest case, just invoke the slice extraction with an index only:

    .. code-block:: yaml

       - kind: processing
         type: SliceExtraction
         properties:
           parameters:
             position: 5

    This will extract the sixth slice (index five) along the first axis (index
    zero).

    If you would like to extract a slice along the second axis (with index
    one), simply provide both parameters, index and axis:

    .. code-block:: yaml

       - kind: processing
         type: SliceExtraction
         properties:
           parameters:
             position: 5
             axis: 1

    This will extract the sixth slice along the second axis.

    And as it is sometimes more convenient to give ranges in axis values
    rather than indices, even this is possible. Suppose the axis you would
    like to extract a slice from runs from 340 to 350 and you would like to
    extract the slice corresponding to 343:

    .. code-block:: yaml

       - kind: processing
         type: SliceExtraction
         properties:
           parameters:
             position: 343
             unit: axis

    In case of you providing the range in axis units rather than indices,
    the value closest to the actual axis value will be chosen automatically.

    For ND datasets with N>2, you can either extract a 1D or ND slice,
    with N always at least one dimension less than the original data. To
    extract a 2D slice from a 3D dataset, simply proceed as above, providing
    one value each for position and axis. If, however, you want to extract a
    1D slice from a 3D dataset, you need to provide two values each for
    position and axis:

    .. code-block:: yaml

       - kind: processing
         type: SliceExtraction
         properties:
           parameters:
             position: [21, 42]
             axis: [0, 2]

    This particular case would be equivalent to ``data[21, :, 42]`` assuming
    ``data`` to contain the numeric data, besides, of course, that the
    processing step takes care of removing the axes as well.

    """


class BaselineCorrection(aspecd.processing.BaselineCorrection):
    """Subtract baseline from dataset.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.BaselineCorrection`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In the simplest case, just invoke the baseline correction with default
    values:

    .. code-block:: yaml

       - kind: processing
         type: BaselineCorrection

    In this case, a zeroth-order polynomial baseline will be subtracted from
    your dataset using ten percent to the left and right, and in case of a
    2D dataset, the baseline correction will be performed along the first
    axis (index zero) for all indices of the second axis (index 1).

    Of course, often you want to control a little bit more how the baseline
    will be corrected. This can be done by explicitly setting some parameters.

    Suppose you want to perform a baseline correction with a polynomial of
    first order:

    .. code-block:: yaml

       - kind: processing
         type: BaselineCorrection
         properties:
           parameters:
             order: 1

    If you want to change the (percental) area used for fitting the
    baseline, and even specify different ranges left and right:

    .. code-block:: yaml

       - kind: processing
         type: BaselineCorrection
         properties:
           parameters:
             fit_area: [5, 20]

    Here, five percent from the left and 20 percent from the right are used.

    Finally, suppose you have a 2D dataset and want to average along the
    second axis (index one):

    .. code-block:: yaml

       - kind: processing
         type: BaselineCorrection
         properties:
           parameters:
             axis: 1

    Of course, you can combine the different options.

    """


class Filtering(aspecd.processing.Filtering):
    """Filter data.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.Filtering`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Generally, filtering requires to provide both, a type of filter and a
    window length. Therefore, for uniform and Gaussian filters, this would be:

    .. code-block:: yaml

       - kind: processing
         type: Filtering
         properties:
           parameters:
             type: uniform
             window_length: 10

    Of course, at least uniform filtering (also known as boxcar or moving
    average) is strongly discouraged due to the artifacts introduced.
    Probably the best bet for applying a filter to smooth your data is the
    Savitzky-Golay filter:

    .. code-block:: yaml

       - kind: processing
         type: Filtering
         properties:
           parameters:
             type: savitzky-golay
             window_length: 10
             order: 3

    Note that for this filter, you need to provide the polynomial order as
    well. To get best results, you will need to experiment with the
    parameters a bit.

    """
